"""
Wisconsin Autonomous - https://www.wisconsinautonomous.org

Copyright (c) 2021 wisconsinautonomous.org
All rights reserved.

Use of this source code is governed by a BSD-style license that can be found
in the LICENSE file at the top level of the repo
"""


class WASystem:
    """Used to manage simulation parameters and can be used to update simulation dynamics

    A system is used for organizational purposes almost exclusively. It is passed to other
    module classes to handle passing various attributes. Could be responsible for updating
    module dynamics depending on the underlying models.

    To instantiate a WASystem, you can either pass in various parameters directly, i.e. ``WASystem(step_size, render_step_size, ...)``
    or pass a ``argparse.Namespace`` in directly. This object is what is returned from :meth:`~WAArgumentParser.parse_args`.

    .. highlight:: python
    .. code:: python

        # Usage
        from wa_simulator import WAArgumentParser, WASystem

        # Passing values directly to the system
        system = WASystem(step_size=1e-3, render_step_size=2e-3, end_time=20)

        # Or using WAArgumentParser
        parser = WAArgumentParser(use_sim_defaults=True)
        args = parser.parse_args()

        system = WASystem(args=args)

    Args:
        step_size (float, optional): the step size for the simulation. Defaults to 5e-2.
        render_step_size (float, optional): Render step size. Defaults to 2e-2.
        end_time (float, optional): the end time for the simulation. Defaults to 120 seconds.
        args (argparse.Namespace, optional): the output namespace from argparse (see explanation above)

    Attributes:
        step_number (int): Counter of the Advance function
        step_size (float): the step size for the simulation
        render_step_size (float): the render step size
        end_time (float): the end time for the simulation
        time (float): stores the time from the simulation
    """

    def __init__(self, step_size: float = 5e-2, render_step_size: float = 2e-2, end_time: float = 120, args: 'argparse.Namespace' = None):

        if args is not None:
            if not all(hasattr(args, attr) for attr in ['step_size', 'render_step_size', 'end_time']):
                raise AttributeError(f"'args' must have 'step_size', 'render_step_size' and 'end_time' members.")
            self.step_size = args.step_size
            self.render_step_size = args.render_step_size
            self.end_time = args.end_time
        else:
            self.step_size = step_size
            self.render_step_size = render_step_size
            self.end_time = end_time

        self.time = 0
        self.step_number = 0

    def advance(self):
        """Advance the system

        Will update the time and increment the step number
        """
        self.time += self.step_size
        self.step_number += 1

    def is_ok(self) -> bool:
        """Check whether the system is still running correctly.

        Simply verifies the end time for the simulation hasn't been reached
        """
        return self.end_time > self.time



"""
Wisconsin Autonomous - https://www.wisconsinautonomous.org

Copyright (c) 2021 wisconsinautonomous.org
All rights reserved.

Use of this source code is governed by a BSD-style license that can be found
in the LICENSE file at the top level of the repo
"""

from abc import ABC, abstractmethod  # Abstract Base Class

# WA Simulator
from wa_simulator.base import WABase
from wa_simulator.core import WA_EARTH_RADIUS, WA_PI, WAVector
from wa_simulator.utils import _load_json, _check_field, _check_field_allowed_values, _WAStaticAttribute, get_wa_data_file

# Other Imports
import json
import numpy as np


def load_sensor_suite_from_json(manager: 'WASensorManager', filename: str):
    """Load a sensor suite from json

    Each loaded sensor will be added to the manager

    Args:
        manager (WASensorManager): The sensor manager to store all created objects in
        filename (str): The json specification file that describes the sensor suite
    """
    j = _load_json(filename)

    # Validate the json file
    _check_field(j, 'Type', value='Sensor')
    _check_field(j, 'Template', value='Sensor Suite')
    _check_field(j, 'Sensors', field_type=list)

    # Load the sensors
    for sensor in j['Sensors']:
        new_sensor = load_sensor_from_json(sensor, manager._system)
        manager.add_sensor(new_sensor)


def load_sensor_from_json(system: 'WASystem', filename: str, **kwargs) -> 'WASensor':
    """Load a sensor from json

    Args:
        system (WASystem): The system for the simulation
        filename (str): The json specification file that describes the sensor
        kwargs: Keyworded arguments. Must contain a 'vehicle' or 'body', not both.

    Returns:
        WASensor: The created sensor
    """
    j = _load_json(filename)

    # Validate the json file
    _check_field(j, 'Type', value='Sensor')
    _check_field(j, 'Template', allowed_values=['IMU', 'GPS'])

    # Check the template and create a sensor based off what it is
    if j['Template'] == 'IMU':
        sensor = WAIMUSensor(system, filename, **kwargs)
    elif j['Template'] == 'GPS':
        sensor = WAGPSSensor(system, filename, **kwargs)
    else:
        raise TypeError(f"{j['Template']} is not an implemented sensor type")

    return sensor


class WASensorManager(WABase):
    """Base class that manages sensors.

    Args:
        system (WASystem): The system for the simulation
        filename (str, optional): A json file to load a scene from. Defaults to None (does nothing).
    """

    # Global filenames for sensor suites
    _EGP_SENSOR_SUITE_FILE = "sensors/suites/ev_grand_prix.json"

    EGP_SENSOR_SUITE_FILE = _WAStaticAttribute('_EGP_SENSOR_SUITE_FILE', get_wa_data_file)

    def __init__(self, system: 'WASystem', filename: str = None):
        self._system = system

        self._sensors = []

        if filename is not None:
            load_sensor_suite_from_json(filename, self)

    def add_sensor(self, sensor: "WASensor"):
        """Add a sensor to the sensor manager"""
        self._sensors.append(sensor)

    def synchronize(self, time):
        """Synchronize each sensor at the specified time

        Args:
            time (float): the time at which the sensors are synchronized to
        """
        for sensor in self._sensors:
            sensor.synchronize(time)

    def advance(self, step):
        """Advance the state of each sensor by the specified time step

        Args:
            step (float): the step to update the sensor by
        """
        for sensor in self._sensors:
            sensor.advance(step)

    def is_ok(self) -> bool:
        for sensor in self._sensors:
            if not sensor.is_ok():
                return False
        return True


class WANoiseModel(ABC):
    """Base class for a noise model

    Very similar to ChNoiseModel (https://github.com/projectchrono/chrono/blob/feature/sensor/src/chrono_sensor/ChNoiseModel.h)
    Made because ChSensor currently _requires_ OptiX (NVIDIA proprietary software)
    """

    @abstractmethod
    def add_noise(data: list):
        """Add noise to the data

        Args:
            data (list): The data to add noise to
        """
        pass


class WANoNoiseModel(WANoiseModel):
    """Derived noise model. Does nothing"""

    def add_noise(data: list):
        """Do nothing"""
        pass


class WANormalDriftNoiseModel(WANoiseModel):
    """Derived noise model. Gaussian drifting noise with noncorrelated equal distributions

    Very similar to ChNormalDriftModel (https://github.com/projectchrono/chrono/blob/feature/sensor/src/chrono_sensor/ChNoiseModel.h)
    Made because ChSensor currently _requires_ OptiX (NVIDIA proprietary software)

    Args:
        p (dict): Model properties/parameters taken from a json specification file
    """

    def __init__(self, p: dict):
        # Validate properties
        _check_field(p, 'Noise Type', value='Normal Drift')
        _check_field(p, 'Update Rate', field_type=float)
        _check_field(p, 'Mean', field_type=list)
        _check_field(p, 'Standard Deviation', field_type=list)
        _check_field(p, 'Bias Drift', field_type=float)
        _check_field(p, 'Tau Drift', field_type=float)

        self._load_properties(p)

    def _load_properties(self, p: dict):
        """Private function that loads propeties and sets them to class variables

        Args:
            p (dict): Propeties dictionary
        """
        self._update_rate = p['Update Rate']
        self._mean = WAVector(p['Mean'])
        self._sigma = WAVector(p['Standard Deviation'])
        self._bias_drift = p['Bias Drift']
        self._tau_drift = p['Tau Drift']

        self._bias = WAVector()

    def add_noise(self, data: list):
        """Add noise to the data

        Args:
            data (list): The data to add noise to
        """

        def rand(element):
            return np.random.normal(getattr(self._mean, element, 0), getattr(self._sigma, element, 0))

        eta_a = WAVector([rand('x'), rand('y'), rand('z')])

        eta_b = WAVector()
        if self._tau_drift > 0 and self._bias_drift > 0:
            def rand(): return np.random.normal(0.0, self._drift_bias * np.sqrt(1.0 / (self._update_rate * self._tau_drift)))  # noqa
            eta_b = WAVector([rand(), rand(), rand()])

        self._bias += eta_b
        data += eta_a + self._bias


class WANormalNoiseModel(WANoiseModel):
    """Derived noise model. Gaussian drifting noise with noncorrelated equal distributions

    Very similar to ChNormalModel (https://github.com/projectchrono/chrono/blob/feature/sensor/src/chrono_sensor/ChNoiseModel.h)
    Made because ChSensor currently _requires_ OptiX (NVIDIA proprietary software)

    Args:
        p (dict): Model properties/parameters taken from a json specification file
    """

    def __init__(self, p: dict):
        # Validate properties
        _check_field(p, 'Noise Type', value='Normal')
        _check_field(p, 'Mean', field_type=list)
        _check_field(p, 'Standard Deviation', field_type=list)

        self._load_properties(p)

    def _load_properties(self, p: dict):
        """Private function that loads propeties and sets them to class variables

        Args:
            p (dict): Propeties dictionary
        """
        self._mean = WAVector(p['Mean'])
        self._sigma = WAVector(p['Standard Deviation'])

    def add_noise(self, data: list):
        """Add noise to the data

        Args:
            data (list): The data to add noise to
        """

        def rand(element):
            return np.random.normal(getattr(self._mean, element, 0), getattr(self._sigma, element, 0))

        data += WAVector([rand('x'), rand('y'), rand('z')])


class WASensor(WABase):
    """Base class for a sensor

    For the WASimulator, a sensor is a component that produces some sensor data. Each sensor is managed by a :class:`~WASensorManager`
    and will typically provide information for a vehicle or the surroundings. Common sensors used in autonomous vehicle
    applications: GPS, IMU, Camera, LiDAR. 

    Sensor implementations are some what limited to the underlying backend for the simulation. If Chrono is used, 
    Physics Based Rendering (PBR) sensors are available through the ray tracing engine `OptiX <https://developer.nvidia.com/optix>`_ 
    and `Chrono::Sensor <https://api.projectchrono.org/group__sensor__sensors.html>`_.

    Args:
        vehicle (WAVehicle): The vehicle to attach to. If not passed, body must be passed.
        body (WABody): The body to attach to. If not passed, vehicle must be passed.
    """

    def __init__(self, vehicle: 'WAVehicle', body: 'WABody'):
        if vehicle is not None and body is not None:
            raise ValueError(f"'vehicle' and 'body' are both set, but this is not allowed. Please only pass one.")
        elif vehicle is None and body is None:
            raise ValueError(f"'vehicle' and 'body' are both not set. Please set one.")

        if vehicle is not None:
            pass
        else:
            if not hasattr(body, 'position') or not isinstance(body.position, WAVector):
                raise ValueError(f"'body' must have a position attribute and it must be a 'WAVector'.")

    @abstractmethod
    def synchronize(self, time: float):
        pass

    @abstractmethod
    def advance(self, step: float):
        pass

    def is_ok(self) -> bool:
        return True

    @abstractmethod
    def get_data(self):
        """Get the sensor data

        Return types may vary. See the respective sensor for specifics.
        """
        pass


class WAIMUSensor(WASensor):
    """Derived sensor class that implements an IMU model

    IMU stands for Inertial Measurement Unit. They come in all shapes and sizes. Typically, an IMU is used to 
    measure position and orientation of whatever it is attached to. To do this, it is actually made up of more sensors,
    which when combined, produce a localizable reading. Common sub-sensors include Gyroscopes, Accelerometers, Altimeters and
    Magnetometer.

    This implementation combines three sensors to produce the expected output: Gyroscope, Accelerometer and Magnetometer. A good
    description of the underyling implementation can be found in `these slides <https://stanford.edu/class/ee267/lectures/lecture9.pdf>`_.

    Args:
        system (WASystem): The system for the simulation
        filename (str): a json specification file that describes an IMU sensor model
        vehicle (WAVehicle, optional): The vehicle to attach to. If not passed, body must be passed.
        body (WABody, optional): The body to attach to. If not passed, vehicle must be passed.
    """

    # Global filenames for imu sensors
    _SBG_IMU_SENSOR_FILE = "sensors/models/SBG_IMU.json"

    SBG_IMU_SENSOR_FILE = _WAStaticAttribute('_SBG_IMU_SENSOR_FILE', get_wa_data_file)

    def __init__(self, system: 'WASystem', filename: str, vehicle: 'WAVehicle' = None, body: 'WABody' = None):
        super().__init__(vehicle, body)

        if body is not None:
            raise NotImplementedError("Setting 'body' is currently not supported. Please pass a vehicle instead")

        self._vehicle = vehicle

        j = _load_json(filename)

        # Validate the json file
        _check_field(j, 'Type', value='Sensor')
        _check_field(j, 'Template', value='IMU')
        _check_field(j, 'Properties', field_type=dict)

        p = j['Properties']
        _check_field(p, 'Update Rate', field_type=int)
        _check_field(p, 'Noise Model', field_type=dict, optional=True)

        if 'Noise Model' in p:
            _check_field(p['Noise Model'], 'Noise Type',
                         allowed_values=["Normal", "Normal Drift"])

        self._load_properties(p)

        self._acc = WAVector()
        self._omega = WAVector()
        self._orientation = WAVector()

    def _load_properties(self, p: dict):
        """Private function that loads properties and sets them to class variables

        Args:
            p (dict): Properties dictionary
        """
        self._update_rate = p['Update Rate']

        self._noise_model = WANoNoiseModel()
        if 'Noise Model' in p:
            n = p['Noise Model']

            if n['Noise Type'] == 'Normal Drift':
                self._noise_model = WANormalDriftNoiseModel(n)
            elif n['Noise Type'] == 'Normal':
                self._noise_model = WANormalNoiseModel(n)
            else:
                raise TypeError(
                    f"{p['Noise Type']} is not an implemented model type")

    def synchronize(self, time):
        """Synchronize the sensor at the specified time

        Args:
            time (float): the time at which the sensors are synchronized to
        """
        self._noise_model.add_noise(self._acc)  # Acceleration (accelerometer)
        # Angular velocity (gyroscope)
        self._noise_model.add_noise(self._omega)
        # self.noise_model.add_noise(self.orientation)  # Orientation ("Magnometer") # noqa

    def advance(self, step):
        """Advance the state of the sensor by the specified time step

        Args:
            step (float): the step to update the sensor by
        """
        self._pos_dtdt = self._vehicle.get_pos_dtdt()
        self._rot = self._vehicle.get_rot()
        self._rot_dt = self._vehicle.get_rot_dt()

    def get_data(self):
        """Get the sensor data

        Returns:
            (WAVector, WAQuaternion, WAQuaternion): Tuple in the form of (acceleration, angular_velocity, orientation)
        """
        return self._pos_dtdt, self._rot_dt, self._rot


class WAGPSSensor(WASensor):
    """Derived sensor class that implements an GPS model

    GPS stands for Global Positioning System. GPS's are everyday sensors you can find in your computer, watch, phone, etc.
    They essentially ping satilites orbiting earth for positional information. This information, in real life, is not very
    precise. IMU's and GPS's are commonly "fused" together to achieve highly accurate and reliable position information.

    Args:
        system (WASystem): The system for the simulation
        filename (str): a json specification file that describes an GPS sensor model
        vehicle (WAVehicle, optional): The vehicle to attach to. If not passed, body must be passed.
        body (WABody, optional): The body to attach to. If not passed, vehicle must be passed.
    """

    # Global filenames for gps sensors
    _SBG_GPS_SENSOR_FILE = "sensors/models/SBG_GPS.json"

    SBG_GPS_SENSOR_FILE = _WAStaticAttribute('_SBG_GPS_SENSOR_FILE', get_wa_data_file)

    def __init__(self, system: 'WASystem', filename: str, vehicle: 'WAVehicle' = None, body: 'WABody' = None):
        super().__init__(vehicle, body)

        if body is not None:
            raise NotImplementedError("Setting 'body' is currently not supported. Please pass a vehicle instead")

        self._vehicle = vehicle

        j = _load_json(filename)

        # Validate the json file
        _check_field(j, 'Type', value='Sensor')
        _check_field(j, 'Template', value='GPS')
        _check_field(j, 'Properties', field_type=dict)

        p = j['Properties']
        _check_field(p, 'Update Rate', field_type=int)
        _check_field(p, 'GPS Reference', field_type=list)
        _check_field(p, 'Noise Model', field_type=dict, optional=True)

        if 'Noise Model' in p:
            _check_field(p['Noise Model'], 'Noise Type',
                         allowed_values=["Normal", "Normal Drift"])

        self._load_properties(p)

    def _load_properties(self, p: dict):
        """Private function that loads properties and sets them to class variables

        Args:
            p (dict): Properties dictionary
        """
        self._update_rate = p['Update Rate']
        self._reference = WAVector(p['GPS Reference'])

        self._noise_model = WANoNoiseModel()
        if 'Noise Model' in p:
            n = p['Noise Model']

            if n['Noise Type'] == 'Normal Drift':
                self._noise_model = WANormalDriftNoiseModel(n)
            elif n['Noise Type'] == 'Normal':
                self._noise_model = WANormalNoiseModel(n)
            else:
                raise TypeError(
                    f"{p['Noise Type']} is not an implemented model type")

        self._pos = WAVector()

    def synchronize(self, time):
        """Synchronize the sensor at the specified time

        Args:
            time (float): the time at which the sensors are synchronized to
        """
        self._noise_model.add_noise(self._pos)

    def advance(self, step):
        """Advance the state of the sensor by the specified time step

        Args:
            step (float): the step to update the sensor by
        """
        self._pos = self._vehicle.get_pos()
        self._coord = WAGPSSensor.cartesian_to_gps(self._pos, self._reference)

    def get_data(self):
        """Get the sensor data

        Returns:
            WAVector: The coordinate location of the vehicle in the form of [longitude, latitude, altitude]
        """
        return self._coord

    @staticmethod
    def cartesian_to_gps(coords: WAVector, ref: WAVector):
        """Convert a point from cartesian to gps

        Args:
          coords(WAVector): The coordinate to convert
          ref(WAVector): The "origin" or reference point

        Returns:
          WAVector: The coordinate in the form of[longitude, latitude, altitude]
        """
        lat = (coords.y / WA_EARTH_RADIUS) * 180.0 / WA_PI + ref.y
        lon = (coords.x / (WA_EARTH_RADIUS * np.cos(lat * WA_PI / 180.0))) * 180.0 / WA_PI + ref.x  # noqa
        alt = coords.z + ref.z

        lon = lon + 360.0 if lon < -180.0 else lon - 360.0 if lon > 180.0 else lon

        return WAVector([lon, lat, alt])

    @staticmethod
    def gps_to_cartesian(coords: WAVector, ref: WAVector):
        """Convert a gps coordinate to cartesian given some reference

        Args:
          coords(WAVector): The coordinate to convert
          ref(WAVector): The "origin" or reference point

        Returns:
          WAVector: The x, y, z point in cartesian
        """
        lon = coords.x
        lat = coords.y
        alt = coords.z

        x = ((lon - ref.x) * WA_PI / 180.0) * (WA_EARTH_RADIUS * np.cos(lat * WA_PI / 180.0))  # noqa
        y = ((lat - ref.y) * WA_PI / 180.0) * WA_EARTH_RADIUS
        z = alt - ref.z

        return WAVector([x, y, z])



"""
Wisconsin Autonomous - https://www.wisconsinautonomous.org

Copyright (c) 2021 wisconsinautonomous.org
All rights reserved.

Use of this source code is governed by a BSD-style license that can be found
in the LICENSE file at the top level of the repo
"""

import unittest
import numpy as np

# Import the track module
from wa_simulator.utils import get_wa_data_file
from wa_simulator.path import WASplinePath, load_waypoints_from_csv, calc_path_length_cummulative, calc_path_curvature
from wa_simulator.track import WATrack, create_constant_width_track

# -----
# Tests
# -----


class TestWATrack(unittest.TestCase):
    """Tests method related to the WATrack"""

    def test_create_constant_width_track(self):
        """Tests the create_constant_width_track method"""
        # Create the centerline
        filename = get_wa_data_file("paths/sample_medium_loop.csv")
        points = load_waypoints_from_csv(filename, delimiter=",")
        path = WASplinePath(points, num_points=1000)

        # Create the track
        track = create_constant_width_track(path, width=6)


if __name__ == '__main__':
    unittest.main()



"""
Wisconsin Autonomous - https://www.wisconsinautonomous.org

Copyright (c) 2021 wisconsinautonomous.org
All rights reserved.

Use of this source code is governed by a BSD-style license that can be found
in the LICENSE file at the top level of the repo
"""

from abc import ABC, abstractmethod  # Abstract Base Class

# WA Simulator
from wa_simulator.base import WABase
from wa_simulator.visualization import WAMatplotlibVisualization
from wa_simulator.utils import _check_type

# Other imports
import sys
import numpy as np


class WAController(WABase):
    """Base class for a controller

    Controllers are responsible for outputting a steering, throttle and braking value.
    This is done because in real life, those are the inputs our cars will have. The
    derived controller's (i.e. the new class that inherits from this class)
    responsibility is to take inputs from the simulation and return these values
    through the get_inputs method.

    Args:
        system (WASystem): The system used to manage the simulation
        vehicle_inputs (WAVehicleInputs): The vehicle inputs

    Attributes:
        steering (float): steering input.
        throttle (float): throttle input.
        braking (float): braking input.
    """

    def __init__(self, system: 'WASystem', vehicle_inputs: 'WAVehicleInputs'):
        self._system = system
        self._vehicle_inputs = vehicle_inputs

    @abstractmethod
    def synchronize(self, time: float):
        pass

    @abstractmethod
    def advance(self, step: float):
        pass

    def is_ok(self) -> bool:
        return True

    def get_inputs(self) -> 'WAVehicleInputs':
        """Get the vehicle inputs

        Returns:
            WAVehicleInputs: The input class
        """
        return self._vehicle_inputs

    def _get_steering(self) -> float:
        return self._vehicle_inputs.steering

    def _get_throttle(self) -> float:
        return self._vehicle_inputs.throttle

    def _get_braking(self) -> float:
        return self._vehicle_inputs.braking

    def _set_steering(self, steering: float):
        self._vehicle_inputs.steering = steering

    def _set_throttle(self, throttle: float):
        self._vehicle_inputs.throttle = throttle

    def _set_braking(self, braking: float):
        self._vehicle_inputs.braking = braking

    steering = property(_get_steering, _set_steering)
    throttle = property(_get_throttle, _set_throttle)
    braking = property(_get_braking, _set_braking)


class _WAKeyboardController(WAController):
    """Base keyboard controller. Still must be inherited (can't be instantiated). Has utilites.

    Args:
        system (WASystem): The system used to manage the simulation
        vehicle_inputs (WAVehicleInputs): The vehicle inputs
    """

    def __init__(self, system: 'WASystem', vehicle_inputs: 'WAVehicleInputs'):
        super().__init__(system, vehicle_inputs)

        self._steering_target = 0
        self._throttle_target = 0
        self._braking_target = 0

        self._steering_delta = system.render_step_size / 2.0
        self._throttle_delta = system.render_step_size / 6.0
        self._braking_delta = system.render_step_size / 1.5

        self._steering_gain = 4.0
        self._throttle_gain = 4.0
        self._braking_gain = 4.0

    def set_steering_delta(self, steering_delta: float):
        """Sets the steering delta value

        Args:
            steering_delta (float): the new steering delta value
        """
        self._steering_delta = steering_delta

    def set_throttle_delta(self, throttle_delta: float):
        """Sets the throttle delta value

        Args:
            throttle_delta (float): the new throttle delta value
        """
        self._throttle_delta = throttle_delta

    def set_braking_delta(self, braking_delta: float):
        """Sets the braking delta value

        Args:
            braking_delta (float): the new braking delta value
        """
        self._braking_delta = braking_delta

    def set_gains(self, steering_gain: float, throttle_gain: float, braking_gain: float):
        """Sets the controllers gains

        Args:
            steering_gain (float): the new steering gain
            throttle_gain (float): the new throttle gain
            braking_gain (float): the new braking gain
        """
        self._steering_gain = steering_gain
        self._throttle_gain = throttle_gain
        self._braking_gain = braking_gain

    def advance(self, step: float):
        """Advance the controller by the specified step

        Integrates dynamics over some step range. If the original step is the same as the passed
        step value, the method is only run once.

        Args:
                step (float): the time step at which the controller should be advanced
        """
        # Integrate dynamics, taking as many steps as required to reach the value 'step'
        t = 0
        while t < step:
            h = step - t

            steering_deriv = self._steering_gain * (self._steering_target - self.steering)
            throttle_deriv = self._throttle_gain * (self._throttle_target - self.throttle)
            braking_deriv = self._braking_gain * (self._braking_target - self.braking)

            self.steering += min(self._steering_delta, h * steering_deriv, key=abs)
            self.throttle += min(self._throttle_delta, h * throttle_deriv, key=abs)
            self.braking += min(self._braking_delta, h * braking_deriv, key=abs)

            t += h

    def _update(self, key: int):
        """Update the target values based on the key.

        The updated target values are based off the delta values for that respective input.
        In this controller, input values are clipped at [-1,1] or [0,1].

        0: throttle increase, braking decreases (up)
        1: adjust steering right (right)
        2: brake increase, throttle decrease (down)
        3: adjust steering left (left)

        Args:
            key (int): the key input
        """
        allowed_keys = [0, 1, 2, 3]

        # Mouse click
        if key == -1:
            return

        # Up
        if key == 0:
            self._throttle_target = np.clip(self._throttle_target + self._throttle_delta, 0.0, +1.0)
            if self._throttle_target > 0:
                self._braking_target = np.clip(self._braking_target - self._braking_delta * 3, 0.0, +1.0)
        # Down
        elif key == 2:
            self._throttle_target = np.clip(self._throttle_target - self._throttle_delta * 3, 0.0, +1.)
            if self._throttle_target <= 0:
                self._braking_target = np.clip(self._braking_target + self._braking_delta, 0.0, +1.0)

        # Right
        elif key == 1:
            self._steering_target = np.clip(self._steering_target + self._steering_delta, -1.0, +1.)

        # Left
        elif key == 3:

            self._steering_target = np.clip(self._steering_target - self._steering_delta, -1.0, +1.)
        else:
            raise ValueError(
                f"Got key type of '{key}'. Was expecting one of the following: {', '.join(allowed_keys)}")


class WAMatplotlibController(_WAKeyboardController):
    """Controls a vehicle via keyboard input from a matplotlib figure

    Will asynchronously change inputs based on user input to the matplotlib window.

    Args:
        system (WASystem): The system used to manage the simulation
        vehicle_inputs (WAVehicleInputs): The vehicle inputs
        vis (WAMatplotlibVisualization): The visualization that holds a matplotlib figure
    """

    def __init__(self, system: 'WASystem', vehicle_inputs: 'WAVehicleInputs', vis: 'WAMatplotlibVisualization'):
        _check_type(vis, WAMatplotlibVisualization, 'vis', 'WAMatplotlibController::__init__')

        super().__init__(system, vehicle_inputs)

        vis.register_key_press_event(self._key_press)

        self._input_dict = {'up': 0, 'right': 1, 'down': 2, 'left': 3}

    def synchronize(self, time):
        """Synchronize the controller at the specified time

        Doesn't do anything since this controller is completely asynchronous

        Args:
                time (float): the time at which the controller should synchronize all depends to
        """
        pass

    def _key_press(self, value):
        if value in self._input_dict.keys():
            self._update(self._input_dict[value])



"""
Wisconsin Autonomous - https://www.wisconsinautonomous.org

Copyright (c) 2021 wisconsinautonomous.org
All rights reserved.

Use of this source code is governed by a BSD-style license that can be found
in the LICENSE file at the top level of the repo
"""

# WA Simulator
from wa_simulator.utils import _check_type, _load_json, _check_field, get_wa_data_file, _WAStaticAttribute
from wa_simulator.sensor import WASensorManager, WASensor, load_sensor_from_json
from wa_simulator.chrono.vehicle import WAChronoVehicle
from wa_simulator.chrono.utils import ChVector_from_list, ChFrame_from_json, get_chrono_data_file, WAVector_to_ChVector
from wa_simulator.chrono.system import WAChronoSystem

# Chrono specific imports
import pychrono as chrono

try:
    import pychrono.sensor as sens
    missing_chrono_sensor = False
except ImportError:
    missing_chrono_sensor = True

    def sens_import_error(function_name: str):
        """Helper function that prints out if a function in this class is used when Chrono::Sensor isn't available for import"""
        msg = f"{function_name}: Requires Chrono::Sensor and was not found. Chrono::Sensor requires building Chrono from source. Consult Aaron Young (aryoung5@wisc.edu) if you have questions."
        raise RuntimeError(msg)


def load_chrono_sensor_scene_from_json(manager: "WAChronoSensorManager", filename: str):
    """Load a chrono sensor scene from a json specification file. A scene may consist of "World" attributes (i.e. lights) or sensors

    Args:
        manager (WASensorManager): The sensor manager to edit the scene of
        filename (str): The json specification file describing the scene
    """
    if missing_chrono_sensor:
        sens_import_error('load_chrono_sensor_scene_from_json')

    j = _load_json(filename)

    # Validate the json file
    _check_field(j, 'Type', value='Sensor')
    _check_field(j, 'Template', value='Chrono Sensor Scene')
    _check_field(j, 'World', field_type=dict, optional=True)
    _check_field(j, 'Sensors', field_type=list, optional=True)
    _check_field(j, 'Objects', field_type=list, optional=True)

    if 'World' in j:
        w = j['World']

        _check_field(w, 'Point Lights', field_type=list)

        # Create the point lights
        # NOTE: Might be optional in the future
        for p in w['Point Lights']:
            _check_field(p, 'Position', field_type=list)
            _check_field(p, 'Color', field_type=list)
            _check_field(p, 'Maximum Range', field_type=float)

            pos = ChVector_from_list(p['Position'], chrono.ChVectorF)
            color = ChVector_from_list(p['Color'], chrono.ChVectorF)
            max_range = p['Maximum Range']

            manager._manager.scene.AddPointLight(pos, color, max_range)

    if 'Sensors' in j:
        s = j['Sensors']

        for sensor in s:
            new_sensor = load_chrono_sensor_from_json(manager._system, sensor)
            manager.add_sensor(new_sensor)


def load_chrono_sensor_from_json(system: 'WAChronoSystem', filename: str, **kwargs) -> 'WAChronoSensor':
    """Load a chrono sensor from json

    If the passed json file isn't a chrono type, it will call the correct method.

    Args:
        system (WAChronoSystem): The system for the simulation
        filename (str): The json specification file that describes the sensor

    Returns:
            WAChronoSensor: The created sensor
    """
    if missing_chrono_sensor:
        sens_import_error('load_chrono_sensor_from_json')

    # Check if the file can be found in the chrono portion of the data folder
    try:
        j = _load_json(filename)
    except FileNotFoundError:
        # File is not chrono specific, try that now
        j = _load_json(get_wa_data_file(filename))
        return load_sensor_from_json(system, filename, **kwargs)

    return WAChronoSensor(system, filename, **kwargs)


class WAChronoSensorManager(WASensorManager):
    """Derived SensorManager class that essentially wraps a ChSensorManager. Used to maintain sensors.

    Args:
        system (WAChronoSystem): The system for the simulation
        vehicle (WAVehicle): The vehicle each sensor is attached to
        filename (str): A json file to load a scene from. Defaults to None (does nothing).
    """

    # Global filenames for environment models
    _EGP_SENSOR_SCENE_FILE = "sensors/scenes/ev_grand_prix.json"

    EGP_SENSOR_SCENE_FILE = _WAStaticAttribute('_EGP_SENSOR_SCENE_FILE', chrono.GetChronoDataFile)

    def __init__(self, system: 'WAChronoSystem', filename: str = None):
        if missing_chrono_sensor:
            sens_import_error('WAChronoSensorManager.__init__')

        _check_type(system, WAChronoSystem, 'system', 'WAChronoSensorManager.__init__')  # noqa

        super().__init__(system)

        self._manager = sens.ChSensorManager(system._system)

        if filename is not None:
            load_chrono_sensor_scene_from_json(self, filename)

    def add_sensor(self, sensor: "WAChronoSensor"):
        """Add a sensor to the sensor manager"""
        if isinstance(sensor, WAChronoSensor):
            self._manager.AddSensor(sensor._sensor)
        elif isinstance(sensor, WASensor):
            super().add_sensor(sensor)
        else:
            raise TypeError(
                f"WAChronoSensorManager.add_sensor: sensor has unknown type '{type(sensor)}'. Must be 'WAChronoSensor' or 'WASensor.")

    def advance(self, step):
        """Advance the state of the sensor by the specified time step

        Args:
            step(float): the step to update the sensor by
        """
        self._manager.Update()

        super().advance(step)


class WAChronoSensor(WASensor):
    """Derived Sensor class that is still abstract that essentially wraps a ChSensor.

    A ChSensor has to be attached to a body that's present in the Chrono world. That body can then be moved
    which then moves the ChSensor. If a vehicle is passed, the sensor is automatically attached to the chassis
    body with some offset. If a vehicle is not passed, a :class:`~WABody` must be passed and the sensor
    will be attached to that object. If both are passed, an exception is raised.

    If body is passed, it must have one attribute: position.

    Args:
        system (WAChronoSystem): The system for the simulation
        filename (str): The json file that describes this sensor
        vehicle (WAChronoVehicle, optional): The vehicle to attach to. If not passed, body must be passed.
        body (WABody, optional): The body to attach to. If not passed, vehicle must be passed.
    """

    # Global filenames for various sensors
    _MONO_CAM_SENSOR_FILE = "sensors/models/generic_monocamera.json"
    _LDMRS_LIDAR_SENSOR_FILE = "sensors/models/ldmrs.json"

    MONO_CAM_SENSOR_FILE = _WAStaticAttribute('_MONO_CAM_SENSOR_FILE', get_chrono_data_file)
    LDMRS_LIDAR_SENSOR_FILE = _WAStaticAttribute('_LDMRS_LIDAR_SENSOR_FILE', get_chrono_data_file)

    def __init__(self, system: 'WAChronoSystem', filename: str, vehicle: 'WAChronoVehicle' = None, body: 'WABody' = None):
        if missing_chrono_sensor:
            sens_import_error('WAChronoSensor.__init__')

        super().__init__(vehicle, body)

        j = _load_json(filename)

        # Validate the json file
        _check_field(j, 'Type', value='Sensor')
        _check_field(j, 'Template', allowed_values=['Camera', 'Lidar', 'IMU', 'GPS'])  # noqa
        _check_field(j, 'Offset Pose', field_type=dict)

        offset_pose = ChFrame_from_json(j['Offset Pose'])

        if vehicle is not None:
            body = vehicle._vehicle.GetChassisBody()
        else:
            asset = body
            body = chrono.ChBody()
            body.SetPos(WAVector_to_ChVector(asset.position))
            body.SetBodyFixed(True)
            system._system.AddBody(body)

            # Create the chrono sensor through the Sensor chrono class
        self._sensor = sens.Sensor.CreateFromJSON(filename, body, offset_pose)  # noqa

    def synchronize(self, time):
        """Synchronize the sensor at the specified time

        Args:
            time (float): the time at which the sensors are synchronized to
        """
        pass

    def advance(self, step):
        """Advance the state of the sensor by the specified time step

        Args:
            step (float): the step to update the sensor by
        """
        pass

    def get_data(self):
        """Get the sensor data

        Returns:
            np.ndarray: The sensor data. Type depends on the actual sensor
        """
        name = self._sensor.GetName().lower()
        if 'camera' in name:
            buff = self._sensor.GetMostRecentRGBA8Buffer()
            return buff.GetRGBA8Data()
        if 'lidar' in name:
            buff = self._sensor.GetMostRecentXYZIBuffer()
            return buff.GetXYZIData()
        else:
            raise TypeError(
                f"WAChronoSensorManager.get_data: can't determine sensor type from name: '{name}'.")



"""
Wisconsin Autonomous - https://www.wisconsinautonomous.org

Copyright (c) 2021 wisconsinautonomous.org
All rights reserved.

Use of this source code is governed by a BSD-style license that can be found
in the LICENSE file at the top level of the repo
"""

from abc import ABC, abstractmethod  # Abstract Base Class


class _WABaseMeta:
    def __new__(mcls, classname, bases, cls_dict):
        cls = super().__new__(mcls, classname, bases, cls_dict)
        for name, member in cls_dict.items():
            if not getattr(member, '__doc__'):
                member.__doc__ = getattr(bases[-1], name).__doc__
        return cls


class WABase(ABC):
    """The base Abstract Base Class for all WAComponents

    In the wa_simulator package, there are defined "components" that are *updatable*. An *updateable*
    component is most likely an entity that has dynamics and must manipulate or interact with the simulation.
    Some examples of *updateable* components: vehicles, sensors, environments/scenes, etc..

    All of these components should/must inherit from this base class. In that way, it forces the description
    of three methods: :meth:`~synchronize`, :meth:`~advance` and :meth:`~is_ok`. Furthermore, the
    :class:`~wa_simulator.simulation.WASimulationManager` class will expect components that inherit from
    :class:`WABase` (otherwise a :code:`TypeError` will be raised).

    .. highlight:: python
    .. code-block:: python

        # Example inhertiance
        # Many components that already inherit from WABase are provided
        # In most use cases for this simulator, simply inheriting from these 
        # classes will be sufficient/recommended

        from wa_simulator import WABase

        class WAVehicle(WABase):
            def synchronize(self, time):
                pass

            def advance(self, step):
                pass

            def is_ok(self):
                # Because this derives from WABase, the is_ok method can be initialized here 
                # and doesn't have to be implemented in any derived classes (i.e. WACar)
                pass

        class WACar(WAVehicle):
            def synchronize(self, time):
                pass

            def advance(self, step):
                pass

    As described above, this class only has two abstract methods: :meth:`~synchronize` and :meth:`~advance`. See their documentation
    for the motivation behind having two update style methods.
    """

    __metaclass__ = _WABaseMeta

    @abstractmethod
    def synchronize(self, time: float):
        """Update the state of this component at the current time.

        The primary reason to decouple the update method into two separate calls (i.e. :meth:`~synchronize` and :meth:`~advance`)
        is to provide flexibility to the user and is essentially semantic. In most simple cases, a user will only need one of the two. 
        Furthermore, can only use :meth:`~advance` if they prefer and just update their own :code:`time` variable. Given the unknown use cases
        for the simulator at the time of writing, it was chosen to provide two different methods with similar functionality as to allow
        the user to choose their desired implementation, rather than the writers of this package.

        As opposed to :meth:`~advance`, this method gets the current time of the simulation. As menthioned earlier, 
        :meth:`~advance` and a user defined `time` variable could be used to instead to hold the current state of the simulation. However,
        to aid in generality of the package, this method is provided to simply provide the current time of the simulation to the user in a decoupled
        manner from the :meth:`~advance` method.

        Args:
            time (float): The current time to synchronize to
        """
        pass

    @abstractmethod
    def advance(self, step: float):
        """Advance the state of this component by the specified time step.

        The primary reason to decouple the update method into two separate calls (i.e. :meth:`~synchronize` and :meth:`~advance`)
        is to provide flexibility to the user and is essentially semantic. In most simple cases, a user will only need one of the two. 
        Furthermore, can only use :meth:`~advance` if they prefer and just update their own :code:`time` variable. Given the unknown use cases
        for the simulator at the time of writing, it was chosen to provide two different methods with similar functionality as to allow
        the user to choose their desired implementation, rather than the writers of this package.

        Args:
            step (float): The step size to advance this component by
        """
        pass

    @abstractmethod
    def is_ok(self) -> bool:
        """Check whether this component is still alive.

        Depending the type of component, a specific element may "fail". For example, when
        a visualization is used that presents a GUI/window, if the user closes that display,
        this would be considered a component death. Therefore, :meth:`~is_ok` should then return
        False.

        Returns:
            bool: True if still alive, false otherwise
        """
        pass



"""
Wisconsin Autonomous - https://www.wisconsinautonomous.org

Copyright (c) 2021 wisconsinautonomous.org
All rights reserved.

Use of this source code is governed by a BSD-style license that can be found
in the LICENSE file at the top level of the repo
"""

# ----------------------
# Data loading utilities
# ----------------------

import pathlib
import contextlib

# Grab the data folder in the root of this repo
_DATA_DIRECTORY = str((pathlib.Path("..") / "data").resolve())


def get_wa_data_file(filename: str) -> str:
    """Get the absolute path for the filename passed relative to the :data:`~DATA_DIRECTORY`.

    .. highlight:: python
    .. code:: python

        # Example Usage
        from wa_simulator import get_wa_data_file

        # By default, the data directory will be set to '<installation path of wa_simulator>/data'
        path = get_wa_data_file('test.json')

        print(path) # -> '<installation path of wa_simulator>/data/test.json'

    Args:
        filename (str): file relative to the data folder to get the absolute path for

    Returns:
        str: the absolute path of the file
    """
    return str(pathlib.Path(_DATA_DIRECTORY) / filename)


def get_wa_data_directory() -> str:
    """Get the data directory.

    Returns:
        str: The data directory
    """
    return _DATA_DIRECTORY


def set_wa_data_directory(path: str):
    """Set the wa data directory.

    .. highlight:: python
    .. code:: python

        # Example Usage
        from wa_simulator import set_wa_data_directory, get_wa_data_directory

        # By default, the data directory will be set to '<installation path of wa_simulator>/data'
        print(get_wa_data_directory()) # -> '<installation path of wa_simulator>/data'

        # Setting it will permenantly reset the data directory
        set_wa_data_directory('/a/random/directory')
        print(get_wa_data_directory()) # -> '/a/random/directory'

    Args:
        path (str): relative (or absolute) path where the data is stored
    """
    global _DATA_DIRECTORY

    _DATA_DIRECTORY = str(pathlib.Path(path).resolve())


@ contextlib.contextmanager
def set_wa_data_directory_temp(path: str):
    """Set wa data directory and yield the result.

    For some helper functions, :meth:`~get_wa_data_file` is called implicitly (inside the method itself).
    This method will temporarily set the data directory for a short period and return to the
    original value.

    .. highlight:: python
    .. code:: python

        # Example
        from wa_simulator import set_wa_data_directory, set_wa_data_directory_temp, create_path_from_json

        # Get the current files absolute (global) path
        from pathlib import Path
        abs_path = Path(__file__).parent.absolute()

        # By default, the data directory will be set to '<installation path of wa_simulator>/data'
        print(get_wa_data_directory()) # -> '<installation path of wa_simulator>/data'

        # Will set the data directory to this files directory
        # 'local_file.json' will be loaded from the current directory
        # create_path_from_json uses get_wa_data_file to grab a csv file, which should also be found
        # in this files current directory
        with wa.set_wa_data_directory_temp(abs_path):
            print(get_wa_data_directory()) # -> '<current location of this file>'

           path = create_path_from_json('local_file.json')

        # the data directory has been restored to its original value
        print(get_wa_data_directory()) # -> '<installation path of wa_simulator>/data'

       # The original data directory will now be reset, so 'global_file.json' should be in
       # the global data directory
       path = create_path_from_json('global_file.json')

    Args:
        path (str): relative (or absolute) path where the data is stored
    """
    global _DATA_DIRECTORY

    _OLD_DATA_DIRECTORY = f'{_DATA_DIRECTORY}'
    set_wa_data_directory(path)
    yield
    _DATA_DIRECTORY = _OLD_DATA_DIRECTORY

# -----------------------
# Type checking utilities
# -----------------------


def _check_type(obj, correct_type, variable_name: str, function_name: str):
    """Check the type of an object to verify it is the correct type

    Args:
        obj (Any): The original object to check the type of
        correct_type (Any): The type obj should be
        variable_name (str): The name of the original variable for printing
        function_name (str): The name of the calling function

    Raises:
        TypeError: if obj is not the right type
    """
    if not isinstance(obj, correct_type):
        raise TypeError(
            f"{function_name}: '{variable_name}' should be a '{correct_type}' but is '{type(obj)}.")

# ----------------------
# JSON related utilities
# ----------------------


def _load_json(filename: str) -> dict:
    """Load a json file

    Will simply use the `json library <https://docs.python.org/3/library/json.html>` and return the
    loaded dictionary.

    Args:
        filename (str): The file to load

    Returns:
        dict: The loaded json file contents in a dictionary form.
    """
    import json

    with open(filename) as f:
        j = json.load(f)

    return j


def _check_field(j: dict, field: str, value=None, field_type=None, allowed_values: list = None, optional: bool = False):
    """Check a field in a dict loaded from json

    Args:
        j (dict): The dictionary loaded via a json file
        field (str): The field to check
        value (Any, optional): Some value field must be
        field_type (Type, optional): The type the field must be
        allowed_values: The allowed values
        optional (bool, optional): Whether the field is optional

    Raises:
        KeyError: If the field is not in j
        ValueError: If the value of j[field] does not equal some value
        TypeError: If the type of j[field] is not the same as field_type
        ValueError: j[field] is not one of the allowed_values
    """

    if field not in j:
        if optional:
            return

        raise KeyError(f"_check_field: '{field}' is not in the passed json")

    if value is not None and j[field] != value:
        raise ValueError(f"_check_field: was expecting '{value}' for '{field}', but got '{j[field]}'.")

    if field_type is not None and not isinstance(j[field], field_type):
        raise TypeError(f"_check_field: was expecting '{field}' to be '{field_type}', but was '{type(j[field])}'.")

    if allowed_values is not None:
        _check_field_allowed_values(j, field, allowed_values)


def _check_field_allowed_values(j: dict, field: str, allowed_values: list):
    """Check a field in a dict loaded from json and check that all elements are allowed

    Args:
        j (dict): The dictionary loaded via a json file
        field (str): The field to check
        allowed_values: The allowed values

    Raises:
        ValueError: j[field] is not one of the allowed_values
    """

    _check_field(j, field)

    # Make sure each value is present in the allowed_values list
    # Ignore dicts cause those are objects, allow parsing of those separately
    if j[field] not in allowed_values and j[field] is not dict:
        raise ValueError(f"_check_field_allowed_values: '{j[field]}' is not an allowed value")

# ---------------------
# Private class helpers
# ---------------------


class _WAStaticAttribute:
    def __init__(self, attr, func=None):
        self._attr = attr
        self._func = func

    def __get__(self, instance, owner):
        if self._func is not None:
            return self._func(getattr(owner, self._attr))
        return getattr(owner, self._attr)



"""
Wisconsin Autonomous - https://www.wisconsinautonomous.org

Copyright (c) 2021 wisconsinautonomous.org
All rights reserved.

Use of this source code is governed by a BSD-style license that can be found
in the LICENSE file at the top level of the repo
"""


class WAVehicleInputs:
    """Object used to hold the inputs to the vehicle model

    The value ranges for the vehicle inputs may vary depending on the
    used vehicle model (i.e. radians vs percentages). This class is not reponsible for
    maintaining such properties, simply should be used for passing values around.

    Args:
        steering (float, optional): steering input. Defaults to 0.0.
        throttle (float, optional): throttle input. Defaults to 0.0.
        braking (float, optional): braking input. Defaults to 0.0.

    Attributes:
        steering (float): steering input. 
        throttle (float): throttle input. 
        braking (float): braking input. 
    """

    def __init__(self, steering: float = 0.0, throttle: float = 0.0, braking: float = 0.0):
        self.steering = steering
        self.throttle = throttle
        self.braking = braking




"""
Wisconsin Autonomous - https://www.wisconsinautonomous.org

Copyright (c) 2021 wisconsinautonomous.org
All rights reserved.

Use of this source code is governed by a BSD-style license that can be found
in the LICENSE file at the top level of the repo
"""

import unittest
import numpy as np

# WA Simulator
import wa_simulator.utils as utils

# -----
# Tests
# -----


class TestWAUtils(unittest.TestCase):
    """Tests various package level things"""

    def test_data_directory(self):
        """Tests the global get_wa_data_directory() variable"""
        import os.path

        self.assertTrue(os.path.isdir(utils.get_wa_data_directory()))
        self.assertTrue(os.path.isfile(utils.get_wa_data_directory() + '/test/test.json'))  # noqa

    def test_get_wa_data_file(self):
        """Tests the get_wa_data_file method"""
        import os.path

        self.assertTrue(os.path.isfile(utils.get_wa_data_file('test/test.json')))  # noqa
        self.assertFalse(os.path.isfile(utils.get_wa_data_file('test/test2.json')))  # noqa

    def test_set_wa_data_directory(self):
        """Tests the set_wa_data_directory method"""
        import os

        old = utils.get_wa_data_directory()

        # Absolute
        utils.set_wa_data_directory("/TEST")
        self.assertTrue(utils.get_wa_data_directory() == "/TEST")
        self.assertTrue(old != "/TEST")

        # Relative
        utils.set_wa_data_directory("TEST")
        self.assertTrue(utils.get_wa_data_directory() == os.path.join(os.getcwd(), "TEST"))  # noqa
        self.assertTrue(utils.get_wa_data_file('test/test.json') == os.path.join(os.getcwd(), "TEST/test/test.json"))  # noqa
        self.assertFalse(utils.get_wa_data_file('test/test.json') == os.path.join(os.getcwd(), "TEST/test/test2.json"))  # noqa

        # Reset
        utils.set_wa_data_directory(old)
        self.assertTrue(utils.get_wa_data_directory() == old)

    def test_set_wa_data_directory_temp(self):
        """Tests the set_wa_data_diretory_temp method"""
        import os

        old = utils.get_wa_data_directory()

        # Absolute
        with utils.set_wa_data_directory_temp("/TEST"):
            self.assertTrue(utils.get_wa_data_directory() == "/TEST")
            self.assertTrue(old != "/TEST")
        self.assertFalse(utils.get_wa_data_directory() == "/TEST")

        with utils.set_wa_data_directory_temp("TEST"):
            self.assertTrue(utils.get_wa_data_directory() == os.path.join(os.getcwd(), "TEST"))  # noqa
            self.assertTrue(utils.get_wa_data_file('test/test.json') == os.path.join(os.getcwd(), "TEST/test/test.json"))  # noqa
            self.assertFalse(utils.get_wa_data_file('test/test.json') == os.path.join(os.getcwd(), "TEST/test/test2.json"))  # noqa
        self.assertTrue(utils.get_wa_data_directory() == old)

    def test_check_type(self):
        """Tests the check_type method"""

        with self.assertRaises(TypeError):
            utils._check_type("TEST", int, "test", "test")
        with self.assertRaises(TypeError):
            utils._check_type("TEST", bool, "test", "test")
        with self.assertRaises(TypeError):
            utils._check_type(True, str, "test", "test")

        # Doesn't raise
        try:
            utils._check_type("TEST", str, "test", "test")
            utils._check_type(True, bool, "test", "test")
        except:
            self.fail("Raise exception unexpectedly!")

    def test_load_json(self):
        """Tests the load_json method"""

        try:
            j = utils._load_json(utils.get_wa_data_file('test/test.json'))
        except:
            self.fail("Raise exception unexpectedly!")

        self.assertTrue('Name' in j)
        self.assertTrue('Type' in j)
        self.assertTrue('Template' in j)
        self.assertTrue('Properties' in j)

        self.assertTrue(j['Name'] == 'Test GPS Sensor Model')

        self.assertTrue('Update Rate' in j['Properties'])
        self.assertTrue(isinstance(j['Properties'], dict))

    def test_check_field(self):
        """Tests the check_field method"""
        j = utils._load_json(utils.get_wa_data_file('test/test.json'))

        try:
            utils._check_field(j, 'Name')
            utils._check_field(j, 'Test', optional=True)
            utils._check_field(j, 'Template', value='GPS')
            utils._check_field(j, 'Properties', field_type=dict)
            utils._check_field(j['Properties']['Noise Model'], 'Noise Type',
                               field_type=str, allowed_values=['Normal', 'Test'])
        except:
            self.fail("Raise exception unexpectedly!")

        with self.assertRaises(KeyError):
            utils._check_field(j, "Test")
        with self.assertRaises(TypeError):
            utils._check_field(j, "Name", field_type=bool)
        with self.assertRaises(ValueError):
            utils._check_field(j, "Name", value='Noise Model')
        with self.assertRaises(ValueError):
            utils._check_field(j, "Name", value='Noise Model', optional=True)

    def test_check_field_allowed_values(self):
        """Tests the check_field_allowed_values method"""
        j = utils._load_json(utils.get_wa_data_file('test/test.json'))

        try:
            utils._check_field_allowed_values(j, 'Name', ['Test GPS Sensor Model', 'Test', 'Type'])  # noqa
        except:
            self.fail("Raise exception unexpectedly!")

        with self.assertRaises(ValueError):
            utils._check_field_allowed_values(j, 'Name', ['Test', 'Type'])  # noqa


if __name__ == '__main__':
    unittest.main()



"""
Wisconsin Autonomous - https://www.wisconsinautonomous.org

Copyright (c) 2021 wisconsinautonomous.org
All rights reserved.

Use of this source code is governed by a BSD-style license that can be found
in the LICENSE file at the top level of the repo
"""

from abc import abstractmethod  # Abstract Base Class

# WA Simulator
from wa_simulator.core import WA_GRAVITY, WAVector, WAQuaternion
from wa_simulator.base import WABase
from wa_simulator.utils import _load_json, _WAStaticAttribute, get_wa_data_file

# Other imports
import numpy as np


def load_properties_from_json(filename: str, prop: str) -> dict:
    """Load a specified property from a json specification file

    Will load a json file and extract the passed property field for use
    by the underyling vehicle object

    Args:
        filename (str): the filename location within the set WA data folder for the json file
        property (str): the property to get. Ex: "Vehicle Properties"

    Raises:
        ValueError: The property field isn't found

    Returns:
        dict: the property field extracted from the json file
    """
    j = _load_json(filename)

    if prop not in j:
        raise ValueError(f"{prop} not found in json.")

    return j[prop]


class WAVehicle(WABase):
    """Base class for a WAVehicle.

    To implement a new vehicle model, override this class. A WAVehicle should interact
    with the terrain/assets/world and take three inputs: steering, throttle, braking.

    Args:
        system (WASystem): the system used to run the simulation
        vehicle_inputs (WAVehicleInputs): The vehicle inputs
        filename (str, optional): Filename to be used for visualization properties
    """

    def __init__(self, system: 'WASystem', vehicle_inputs: 'WAVehicleInputs', filename=None):
        self._system = system
        self._vehicle_inputs = vehicle_inputs

        self._vis_properties = (
            dict()
            if filename is None
            else load_properties_from_json(filename, "Visualization Properties")
        )

    def get_visual_properties(self) -> dict:
        """Get visual properties for visualizing the vehicle

        .. todo::
            Define the visual properties required

        Returns:
            dict: the visual properties
        """
        return self._vis_properties

    @abstractmethod
    def synchronize(self, time: float):
        pass

    @abstractmethod
    def advance(self, step: float):
        pass

    def is_ok(self) -> bool:
        return True

    @abstractmethod
    def get_pos(self) -> WAVector:
        """Get the center of mass (COM) position of the vehicle.

        Returns:
            WAVector: the position of the vehicle
        """
        pass

    @abstractmethod
    def get_rot(self) -> WAQuaternion:
        """Get the rotation about the center of mass (COM) of the vehicle

        Returns:
            WAQuaternion: the vehicles orientation
        """
        pass

    @abstractmethod
    def get_pos_dt(self) -> WAVector:
        """Get the instantaneous velocity of the vehicle

        Returns:
            WAVector: The velocity where X is forward, Z is up and Y is left (ISO standard)
        """
        pass

    @abstractmethod
    def get_rot_dt(self) -> WAQuaternion:
        """Get the angular velocity of the vehicle

        Returns:
            WAQuaternion: The angular velocity
        """
        pass

    @abstractmethod
    def get_pos_dtdt(self) -> WAVector:
        """Get the acceleration of the vehicle

        Returns:
            WAVector: The acceleration where X is forward, Z is up and Y is left (ISO standard)
        """
        pass

    @abstractmethod
    def get_rot_dtdt(self) -> WAQuaternion:
        """Get the angular acceleration of the vehicle

        Returns:
            WAQuaternion: The angular acceleration
        """
        pass


class WALinearKinematicBicycle(WAVehicle):
    """A linear, kinematic bicycle model

    Args:
        system (WASystem): the system used to run the simulation
        vehicle_inputs (WAVehicleInputs): The vehicle inputs
        filename (str): json specification file used to describe the parameters for the vehicle
        init_pos (WAVector): initial position
        init_rot (WAQuaternion): initial rotation
        init_pos_dt (WAVector): initial velocity
    """

    # Global filenames for vehicle models
    _GO_KART_MODEL_FILE = "vehicles/GoKart/GoKart_KinematicBicycle.json"

    GO_KART_MODEL_FILE = _WAStaticAttribute('_GO_KART_MODEL_FILE', get_wa_data_file)

    def __init__(self, system: 'WASystem', vehicle_inputs: 'WAVehicleInputs', filename: str, init_pos: WAVector = WAVector(), init_rot: WAQuaternion = WAQuaternion.from_z_rotation(0), init_pos_dt: WAVector = WAVector()):
        super().__init__(system, vehicle_inputs, filename)

        # Simple state variables
        self._x = init_pos.x
        self._y = init_pos.y
        self._yaw = init_rot.to_euler_yaw()
        self._v = init_pos_dt.length
        self._acc = 0.0

        # Wheel angular velocity
        self._omega = 0.0
        self._omega_dot = 0.0

        self._yaw_dt = 0.0
        self._yaw_dtdt = 0.0
        self._last_yaw = 0.0

        properties = load_properties_from_json(filename, "Vehicle Properties")
        self._initialize(properties)

    def _initialize(self, vp):
        """Initialize the vehicle parameters

        Expected properties:
            "Wheelbase" (float): distance between the front and back tire
            "Mass" (float): mass of the vehicle
            "Gear Ratio" (float): gear ratio of the simulated powertrain
            "Effective Radius" (float): 
            "Inertia" (float): Effective inertia for the vehicle
            "Aerodynamic Coefficient" (float):
            "Friction Coefficient" (float):
            "C" (float)
            "Max Force" (float)
            "Torque Coefficients" (list)
            "Steering" ([min, max]): Min and max throttle values
            "Throttle" ([min, max]): Min and max throttle values
            "Braking" ([min, max]): Min and max braking values

        Args:
            vp (dict): Vehicle preoperties (see above for expected keys)
        """

        # Initialize vehicle properties
        self._mass = vp["Mass"]

        # Torque coefficients
        self._a = vp["Torque Coefficients"]

        # Gear ratio, effective radius and inertia
        self._GR = vp["Gear Ratio"]
        self._r_eff = vp["Effective Radius"]
        self._J_e = vp["Inertia"]

        # Aerodynamic and friction coefficients
        self._c_a = vp["Aerodynamic Coefficient"]
        self._c_rl = vp["Friction Coefficient"]

        # Tire forces
        self._c = vp["C"]
        self._F_max = vp["Max Force"]

        self._L = vp["Wheelbase"]
        (self._min_steering, self._max_steering) = vp["Steering"]
        (self._min_throttle, self._max_throttle) = vp["Throttle"]
        (self._min_braking, self._max_braking) = vp["Braking"]

    def synchronize(self, time):
        s = self._vehicle_inputs.steering
        t = self._vehicle_inputs.throttle
        b = self._vehicle_inputs.braking

        self._steering = np.clip(s, self._min_steering, self._max_steering)
        self._throttle = np.clip(t, self._min_throttle, self._max_throttle)
        self._braking = np.clip(b, self._min_braking, self._max_braking)

        if self._braking > 1e-1:
            self._throttle = -self._braking

    def advance(self, step):
        # TODO: COMMENT!
        if self._v == 0 and self._throttle == 0:
            F_x = 0
            F_load = 0
            T_e = 0
        else:
            if self._v == 0:
                # Remove possibity of divide by zero error
                self._v = 1e-8
                self._omega = 1e-8

            # Update engine and tire forces
            F_aero = self._c_a * self._v ** 2
            R_x = self._c_rl * self._v
            F_g = self._mass * WA_GRAVITY * np.sin(0)
            F_load = F_aero + R_x + F_g
            T_e = self._throttle * (
                self._a[0] + self._a[1] * self._omega +
                self._a[2] * self._omega ** 2
            )
            omega_w = self._GR * self._omega  # Wheel angular velocity
            s = (omega_w * self._r_eff - self._v) / self._v  # slip ratio
            cs = self._c * s

            if abs(s) < 1:
                F_x = cs
            else:
                F_x = self._F_max

        if self._throttle < 0:
            if self._v <= 0:
                F_x = 0
            else:
                F_x = -abs(F_x)

        # Update state information
        self._x += self._v * np.cos(self._yaw) * step
        self._y += self._v * np.sin(self._yaw) * step
        self._yaw += self._v / self._L * np.tan(self._steering) * step
        self._v += self._acc * step
        self._acc = (F_x - F_load) / self._mass
        self._omega += self._omega_dot * step
        self._omega_dot = (T_e - self._GR * self._r_eff * F_load) / self._J_e

        v_epsilon = abs(self._v) > 0.1
        self._yaw_dt = (self._last_yaw - self._yaw) / step if v_epsilon else 0
        self._yaw_dtdt = (self._last_yaw - self._yaw) / step if v_epsilon else 0  # noqa
        self._last_yaw = self._yaw

    def get_pos(self) -> WAVector:
        return WAVector([self._x, self._y, 0.0])

    def get_rot(self) -> WAQuaternion:
        return WAQuaternion.from_z_rotation(self._yaw)

    def get_pos_dt(self) -> WAVector:
        return WAVector([self._v * np.cos(self._yaw), self._v * np.sin(self._yaw), 0.0])

    def get_rot_dt(self) -> WAQuaternion:
        return WAQuaternion.from_z_rotation(self._yaw_dt)

    def get_pos_dtdt(self) -> WAVector:
        angle = np.tan(np.interp(self._steering, [-1, 1], [self._min_steering, self._max_steering]))  # noqa
        angle = angle if angle != 0 else 1e-3
        tr = self._L / angle
        tr = tr if tr != 0 else 1e-3
        return WAVector([self._acc, self._v ** 2 / tr, WA_GRAVITY])

    def get_rot_dtdt(self) -> WAQuaternion:
        return WAQuaternion.from_z_rotation(self._yaw_dtdt)



"""
Wisconsin Autonomous - https://www.wisconsinautonomous.org

Copyright (c) 2021 wisconsinautonomous.org
All rights reserved.

Use of this source code is governed by a BSD-style license that can be found
in the LICENSE file at the top level of the repo
"""

from abc import abstractmethod  # Abstract Base Class

# WA Simulator
from wa_simulator.base import WABase
from wa_simulator.core import WAVector
from wa_simulator.track import create_track_from_json
from wa_simulator.utils import _load_json, _check_field, get_wa_data_file, _WAStaticAttribute

# Other imports
import dill


def load_environment_from_json(environment: 'WAEnvironment', filename: str):
    j = _load_json(filename)

    # Validate the json file
    _check_field(j, 'Type', value='Environment')
    _check_field(j, 'Template')
    _check_field(j, 'World', field_type=dict, optional=True)
    _check_field(j, 'Objects', field_type=list, optional=True)

    if 'World' in j:
        w = j['World']

    if 'Objects' in j:
        objects = j['Objects']

        for o in objects:
            _check_field(o, 'Size', field_type=list)
            _check_field(o, 'Position', field_type=list, optional=True)
            _check_field(o, 'Color', field_type=list, optional=True)

            kwargs = {}
            kwargs['size'] = WAVector(o['Size'])
            kwargs['position'] = WAVector(o['Position'])

            if 'Color' in o:
                kwargs['color'] = WAVector(o['Color'])

            if 'Texture' in o:
                kwargs['texture'] = o['Texture']

            if 'Name' in o:
                kwargs['name'] = o['Name']

            environment.add_asset(WABody(**kwargs))

    if 'Track' in j:
        t = j['Track']

        # Validate json
        _check_field(t, 'Track Input File', field_type=str)

        track_file = t['Track Input File']
        track = create_track_from_json(get_wa_data_file(track_file), environment)

        environment.add_asset(track.center)


class WABody:
    """Base class for arbitrary objects in a simulation world that can be visualized.

    .. highlight:: python
    .. code-block:: python

        # Example usage

        from wa_simulator import WABody, WAVector

        size = WAVector()
        pos = WAVector()

        body= WABody(pos=pos, size=size)
        print(body.size, body.pos)


    Args:
        **kwargs: Keyworded arguments that are stored as attributes in this object.
    """

    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)

    def getattr(self, attr: str):
        """Gets the passed attr if present and returns it, returns None if it's not present

        Args:
            attr (str): The attr to get and return

        Returns:
            Any: Returns either None if attr is not present or the actual attribute
        """
        try:
            return getattr(self, attr)
        except AttributeError:
            return None

    def __getstate__(self):
        """Return only the pickleable objects"""
        attrs = {k: v for k, v in self.__dict__.items() if dill.pickles(v)}
        return attrs


class WAEnvironment(WABase):
    """Base class for the environment object.

    The environment object is responsible for handling data or assets within the world.
    Assets may be obstacles, miscellaneous objects, other vehicles or weather.
    """

    def __init__(self):
        self._assets = list()
        self._asset_dict = dict()
        self._updating_assets = list()

    @ abstractmethod
    def synchronize(self, time: float):
        pass

    @ abstractmethod
    def advance(self, step: float):
        pass

    def is_ok(self) -> bool:
        return True

    def create_body(self, **kwargs) -> WABody:
        """Create a body and add it as an asset to the world. Adds the asset using :meth:`~add_asset`.

        Args:
            **kwargs: Keyworded arguments used to create the asset

        Returns:
            WABody: The created body
        """
        body = WABody(**kwargs)
        self.add_asset(body)
        return body

    def add_asset(self, asset: 'Any'):
        """Add an asset to the world

        A asset holds certain attributes that may be applicable in various simulations. This method
        stores created assets in the environment object.

        Args:
            asset (Any): The asset to add.
        """
        self._assets.append(asset)
        if hasattr(asset, 'name'):
            self._asset_dict[asset.name] = asset
        if hasattr(asset, 'updates') and asset.updates:
            self._updating_assets.append(asset)

    def get_assets(self) -> list:
        """Returns the assets that have been attached to this environment.

        Returns:
            list: A list of the assets added through :meth:`~add_asset`.
        """
        return self._assets

    def get_asset(self, name: str) -> 'Any':
        """Get the asset with the attached name

        Args:
            name (str): The name of the asset

        Returns:
            Any: The asset with that name

        Raises:
            KeyError: If the name is not found
        """
        if name not in self._asset_dict:
            raise KeyError(f"Asset with name '{name}' was not found.")

        return self._asset_dict[name]

    def get_updating_assets(self) -> list:
        """Get the list of updating assets

        Updating assets are assets that have an 'update' member variable set to true. Visualizations then update the position of these objects
        throughout the sim.

        Returns:
            list: the list of updating assets
        """
        return self._updating_assets


class WASimpleEnvironment(WAEnvironment):
    """Simple environment.

    Args:
        system (WASystem): The system used to handle simulation meta data
        filename (str): The json specification file used to generate the terrain
    """

    _EGP_ENV_MODEL_FILE = "environments/ev_grand_prix.json"

    EGP_ENV_MODEL_FILE = _WAStaticAttribute('_EGP_ENV_MODEL_FILE', get_wa_data_file)

    def __init__(self, system: 'WASystem', filename: str):
        super().__init__()

        self._system = system

        load_environment_from_json(self, filename)

    def synchronize(self, time: float):
        """Synchronize the environment with the rest of the world at the specified time

        Simple environment doesn't actually do anything for now.

        Args:
            time(float): the time at which the enviornment should be synchronized to
        """
        pass

    def advance(self, step: float):
        """Advance the state of the environment

        Simple environment doesn't actually do anything for now.

        Args:
            step(float): the time step at which the enviornment should be advanced
        """
        pass



from time import time as get_wall_time
import signal
from ._import import _import, _get_dirs, _get_files
from ._version import __version__

__author__ = "Wisconsin Autonomous (wisconsinautonomous@studentorg.wisc.edu)"
"""Wisconsin Autonomous (wisconsinautonomous@studentorg.wisc.edu)"""
__license__ = "BSD3"
"""BSD3"""
__version__ = __version__

for d in _get_dirs(__file__, ignore=["chrono", "data"]):
    _import(d, globals())

for f in _get_files(__file__):
    _import(f, globals())


def _signal_handler(sig, frame):
    """Signal handler that will exit if ctrl+c is recorded in the terminal window.

    Allows easier exiting of a matplotlib plot

    Args:
        sig (int): Signal number
        frame (int): ?
    """

    import sys
    sys.exit(0)


# setup the signal listener to listen for the interrupt signal (ctrl+c)
signal.signal(signal.SIGINT, _signal_handler)

del _import, _get_dirs, _get_files, signal



"""
Wisconsin Autonomous - https://www.wisconsinautonomous.org

Copyright (c) 2021 wisconsinautonomous.org
All rights reserved.

Use of this source code is governed by a BSD-style license that can be found
in the LICENSE file at the top level of the repo
"""

# WA Simulator
from wa_simulator.core import WAVector, WA_PI
from wa_simulator.utils import get_wa_data_file
from wa_simulator.path import WAPath
from wa_simulator.track import WATrack
from wa_simulator.visualization import WAVisualization, WABody
from wa_simulator.vehicle_inputs import WAVehicleInputs
from wa_simulator.chrono.utils import ChVector_to_WAVector, WAVector_to_ChVector

# Chrono specific imports
import pychrono as chrono
import pychrono.vehicle as veh
import pychrono.irrlicht as irr

# Other imports
import numpy as np
from math import ceil


def draw_path_in_irrlicht(system: 'WAChronoSystem', path: 'WAPath'):
    """Draw a WAPath representation as a ChBezierCurve in irrlicht

    Basically just copies over the path's points into something viewable for Chrono

    Args:
        system (WAChronoSystem): system that manages the simulation
        path (WAPath): WA path object to visualize in irrlicht
    """
    points = path.get_points()
    ch_points = chrono.vector_ChVectorD()
    for p in points:
        ch_points.push_back(chrono.ChVectorD(p[0], p[1], p[2]))
    curve = chrono.ChBezierCurve(ch_points)

    road = system._system.NewBody()
    road.SetBodyFixed(True)
    system._system.AddBody(road)

    num_points = len(points)
    path_asset = chrono.ChLineShape()
    path_asset.SetLineGeometry(chrono.ChLineBezier(curve))
    path_asset.SetColor(chrono.ChColor(0, 0.8, 0))
    path_asset.SetNumRenderPoints(max(2 * num_points, 400))
    road.AddAsset(path_asset)


def create_sphere_in_chrono(system: 'WAChronoSystem', rgb=(1, 0, 0)) -> chrono.ChBodyEasySphere:
    """Create and add a sphere to the chrono world 

    Args:
        system (WASystem): the system that manages the simulation
        rgb (tuple, optional): (red, green, blue). Defaults to (1, 0, 0) or red.

    Returns:
        chrono.ChBodyEasySphere: the sphere that was created
    """
    sphere = chrono.ChBodyEasySphere(0.25, 1000, True, False)
    sphere.SetBodyFixed(True)
    sphere.AddAsset(chrono.ChColorAsset(*rgb))

    texture = chrono.ChVisualMaterial()
    texture.SetDiffuseColor(chrono.ChVectorF(*rgb))
    chrono.CastToChVisualization(sphere.GetAssets()[0]).material_list.append(texture)

    system._system.Add(sphere)

    return sphere


def update_position_of_sphere(sphere: chrono.ChBodyEasySphere, pos: WAVector):
    """Update the position of a sphere being visualized in irrlicht

    Args:
        sphere (chrono.ChBodyEasySphere): the sphere to change the position of
        pos (WAVector): the new position
    """
    if not isinstance(pos, WAVector):
        pos = WAVector(pos)

    sphere.SetPos(chrono.ChVectorD(pos.x, pos.y, 0))


class WAChronoIrrlicht(WAVisualization):
    """Chrono irrlicht visualization wrapper. Sets default values for a Vehicle irr app

    Args:
        system (WAChronoSystem): holds information regarding the simulation and performs dynamic updates
        vehicle (WAChronoVehicle): vehicle that holds the Chrono vehicle
        vehicle_inputs (WAVehicleInputs): the vehicle inputs
        environment (WAEnvironment, optional): An environment with various world assets to visualize. Defaults to None (doesn't visualize anything).
        opponents (list, optional): Opponents present in the simulation. The camera will track :attr:`~vehicle` not any opponent.
        record (bool, optional): If set to true, images will be saved under record_filename. Defaults to False (doesn't save images).
        record_folder (str, optional): The folder to save images to. Defaults to "OUTPUT/".
        should_bind (bool, optional): After `all` assets have been added to the environment, irrlicht needs to "bind" and add the assets to the 3D rendered world. This can be done either on instantiation (True) or on the first update (False). Defaults to True.
    """

    def __init__(self, system: 'WAChronoSystem', vehicle: 'WAChronoVehicle', vehicle_inputs: 'WAVehicleInputs', environment: 'WAEnvironment' = None, opponents: list = [], record: bool = False, record_folder: str = "OUTPUT/", should_bind: bool = True):
        self._render_steps = int(
            ceil(system.render_step_size / system.step_size))

        self._system = system
        self._vehicle_inputs = vehicle_inputs

        self._record = record
        self._record_folder = record_folder
        self._save_number = 0

        self._app = veh.ChVehicleIrrApp(vehicle._vehicle)
        self._app.SetHUDLocation(500, 20)
        self._app.SetSkyBox()
        self._app.AddTypicalLogo()
        self._app.AddTypicalLights(
            irr.vector3df(-150.0, -150.0, 200.0),
            irr.vector3df(-150.0, 150.0, 200.0),
            100,
            100,
        )
        self._app.AddTypicalLights(
            irr.vector3df(150.0, -150.0, 200.0),
            irr.vector3df(150.0, 150.0, 200.0),
            100,
            100,
        )
        self._app.SetChaseCamera(chrono.ChVectorD(0.0, 0.0, 1.75), 6.0, 0.5)
        self._app.SetTimestep(system.step_size)

        self._first = True
        if should_bind:
            self.bind()

        if environment is not None:
            self.visualize(environment.get_assets())

    def synchronize(self, time: float):
        """Synchronize the irrlicht app with the vehicle inputs at the passed time

        Args:
            time (double): time to synchronize the simulation to
        """
        d = veh.Inputs()
        d.m_steering = self._vehicle_inputs.steering
        d.m_throttle = self._vehicle_inputs.throttle
        d.m_braking = self._vehicle_inputs.braking

        self._app.Synchronize("", d)

    def advance(self, step: float):
        """Advance the state of the visualization by the specified step

        Will update the render only if the simulation step is a multiple of render steps

        Args:
            step (double): step size to update the visualization by
        """
        if self._first:
            self.bind()

        if self._system.step_number % self._render_steps == 0:
            self._app.BeginScene(True, True, irr.SColor(255, 140, 161, 192))
            self._app.DrawAll()
            self._app.EndScene()

            if self._record:
                self._app.WriteImageToFile(f"{self._record_folder}{self._save_number}.png")
                self._save_number += 1

        self._app.Advance(step)

    def is_ok(self):
        """Checks if the irrlicht rending window is still alive

        Returns:
            bool: whether the simulation is still alive
        """
        return self._app.GetDevice().run()

    def bind(self):
        """Call AssetBindAll and AssetUpdateAll."""
        self._app.AssetBindAll()
        self._app.AssetUpdateAll()

        self._first = False

    def visualize(self, assets, *args, **kwargs):
        for i, asset in enumerate(assets):
            if isinstance(asset, WAPath):
                draw_path_in_irrlicht(self._system, asset)
            elif isinstance(asset, WATrack):
                draw_path_in_irrlicht(self._system, asset.center)
                draw_path_in_irrlicht(self._system, asset.left)
                draw_path_in_irrlicht(self._system, asset.right)


try:
    from wa_simulator.chrono.sensor import WAChronoSensorManager

    import pychrono.sensor as sens

    class WAChronoSensorVisualization(WAVisualization):
        """Chrono sensor visualization wrapper. Sets default values for a camera sensor

        Args:
            system (WAChronoSystem): holds information regarding the simulation and performs dynamic updates
            vehicle (WAChronoVehicle): vehicle that holds the Chrono vehicle
            vehicle_inputs (WAVehicleInputs): the vehicle inputs
            environment (WAEnvironment, optional): An environment with various world assets to visualize. Defaults to None (doesn't visualize anything).
            opponents (list, optional): Opponents present in the simulation. The camera will track :attr:`~vehicle` not any opponent.
            record (bool, optional): If set to true, images will be saved under record_filename. Defaults to False (doesn't save images).
            record_folder (str, optional): The folder to save images to. Defaults to "OUTPUT/".
        """

        def __init__(self, system: 'WAChronoSystem', vehicle: 'WAChronoVehicle', vehicle_inputs: 'WAVehicleInputs', environment: 'WAEnvironment' = None, opponents: list = [], record: bool = False, record_folder: str = "OUTPUT/"):
            self._render_steps = int(
                ceil(system.render_step_size / system.step_size))

            self._system = system
            self._vehicle_inputs = vehicle_inputs

            self._record = record
            self._record_folder = record_folder

            self._manager = WAChronoSensorManager(system)
            self._manager._manager.scene.AddPointLight(chrono.ChVectorF(0, 0, 100), chrono.ChVectorF(2, 2, 2), 5000)

            update_rate = 30.
            offset_pose = chrono.ChFrameD(chrono.ChVectorD(-8, 0, 2))
            image_width = 1280
            image_height = 720
            fov = 1.408
            cam = sens.ChCameraSensor(
                vehicle._vehicle.GetChassisBody(),              # body camera is attached to
                update_rate,            # update rate in Hz
                offset_pose,            # offset pose
                image_width,            # image width
                image_height,           # image height
                fov                    # camera's horizontal field of view
            )

            cam.PushFilter(sens.ChFilterVisualize(image_width, image_height))
            if self._record:
                cam.PushFilter(sens.ChFilterSave(self._record_folder))

            self._manager._manager.AddSensor(cam)

            if environment is not None:
                self.visualize(environment.get_assets())

        def synchronize(self, time: float):
            self._manager.synchronize(time)

        def advance(self, step: float):
            self._manager.advance(step)

        def is_ok(self):
            return True

        def visualize(self, assets, *args, **kwargs):
            pass
except:
    pass



"""
Wisconsin Autonomous - https://www.wisconsinautonomous.org

Copyright (c) 2021 wisconsinautonomous.org
All rights reserved.

Use of this source code is governed by a BSD-style license that can be found
in the LICENSE file at the top level of the repo
"""

import unittest
import numpy as np

# Import the core module
from wa_simulator.core import WAVector, WAQuaternion, WAArgumentParser, WA_PI

# -----
# Tests
# -----


class TestWASimulator(unittest.TestCase):
    """Tests various package level things"""

    def test_main_import(self):
        """Verifies the main import works"""
        import wa_simulator as wa

        self.assertTrue(hasattr(wa, 'WABase'))  # base.py
        self.assertTrue(hasattr(wa, 'WA_PI'))  # constants.py
        self.assertTrue(hasattr(wa, 'WAController'))  # controller.py
        self.assertTrue(hasattr(wa, 'WAVector'))  # core.py
        self.assertTrue(hasattr(wa, 'WAEnvironment'))  # environment.py
        self.assertTrue(hasattr(wa, 'WAVehicleInputs'))  # inputs.py
        self.assertTrue(hasattr(wa, 'WAPath'))  # path.py
        self.assertTrue(hasattr(wa, 'WASensor'))  # sensor.py
        self.assertTrue(hasattr(wa, 'WASimulationManager'))  # simulation.py
        self.assertTrue(hasattr(wa, 'WASystem'))  # system.py
        self.assertTrue(hasattr(wa, 'WATrack'))  # track.py
        self.assertTrue(hasattr(wa, 'get_wa_data_file'))  # utils.py
        self.assertTrue(hasattr(wa, 'WAVehicle'))  # vehicle.py
        self.assertTrue(hasattr(wa, 'WAVisualization'))  # visualization.py


class TestWAVector(unittest.TestCase):
    """Tests methods related to WAVectors"""

    def test_add(self):
        """Tests simple addition of two WAVectors"""
        v1 = WAVector()
        v2 = WAVector([1, 1, 1])

        self.assertEqual(v1 + v2, v2)

    def test_add2(self):
        """Tests simple addition of two WAVectors"""
        v1 = WAVector([91, 44, -10])
        v2 = WAVector([12, -111, 0])

        self.assertEqual(v1 + v2, WAVector([103, -67, -10]))

    def test_add3(self):
        """Tests adding a constant to a WAVector"""
        v = WAVector([1, 1, 1])
        n = 5

        self.assertEqual(v + n, WAVector([6, 6, 6]))

    def test_sub(self):
        """Tests simple subtraction of two WAVectors"""
        v1 = WAVector([11, 5, 10])
        v2 = WAVector([1, 1, 1])

        self.assertEqual(v1 - v2, WAVector([10, 4, 9]))

    def test_sub2(self):
        """Tests subtracting a constant to a WAVector"""
        v = WAVector([1, 1, 1])
        n = 5

        self.assertEqual(v - n, WAVector([-4, -4, -4]))

    def test_mul(self):
        """Tests multiplying a vector by a constant"""
        v = WAVector([1, 2, 3])
        n = 5

        self.assertEqual(v * n, WAVector([5, 10, 15]))

    def test_cross(self):
        """Tests cross product of two WAVectors"""
        v1 = WAVector([1, 4, 5])
        v2 = WAVector([6, 9, 10])

        self.assertEqual(v1.cross(v2), WAVector([-5, 20, -15]))

    def test_dot(self):
        """Tests dot product of two WAVectors"""
        v1 = WAVector([1, 4, 5])
        v2 = WAVector([6, 9, 10])

        self.assertEqual(v1.dot(v2), 92)

    def test_length(self):
        """Tests the length method of a WAVector"""
        v = WAVector([10, -4, 1])

        self.assertEqual(v.length, (10**2 + (-4)**2 + 1**2)**(1/2))


class TestWAQuaternion(unittest.TestCase):
    """Tests methods related to WAQuaternion's"""

    def test_length(self):
        """Tests the length method of a WAQuaternion"""
        q = WAQuaternion([5, 10, -4, 1])

        self.assertEqual(q.length, (5**2 + 10**2 + (-4)**2 + 1**2)**(1/2))

    def test_rot(self):
        """Tests rotation of a vector by a quaternion"""
        q = WAQuaternion.from_z_rotation(WA_PI)  # 180 degree rotation
        v = WAVector([1, 0, 0])

        self.assertTrue(np.allclose(q * v, WAVector([-1, 0, 0])))


class TestWAArgumentParser(unittest.TestCase):
    """Tests method related to WAArgumentParser"""

    def test_default(self):
        """Tests default arguments"""
        cli = WAArgumentParser(True)
        args = cli.parse_args('')

        self.assertEqual(args.step_size, 3e-3)
        self.assertEqual(args.render_step_size, 1 / 10.0)
        self.assertEqual(args.irrlicht, False)
        self.assertEqual(args.matplotlib, False)


if __name__ == '__main__':
    unittest.main()



"""
Wisconsin Autonomous - https://www.wisconsinautonomous.org

Copyright (c) 2021 wisconsinautonomous.org
All rights reserved.

Use of this source code is governed by a BSD-style license that can be found
in the LICENSE file at the top level of the repo
"""

from abc import ABC, abstractmethod  # Abstract Base Class

# WA Simulator
from wa_simulator.base import WABase
from wa_simulator.core import WA_PI, WAVector
from wa_simulator.path import WAPath
from wa_simulator.track import WATrack
from wa_simulator.environment import WABody

# Other imports
from multiprocessing import Process, Pipe, Barrier, Queue, set_start_method
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.patches import Polygon, Patch
import matplotlib.collections as collections
from matplotlib.animation import FuncAnimation
from IPython.display import display, clear_output


class WAVisualization(WABase):
    """Base class to be used for visualization of the simulation world.

    Inherits from WABase, so this class and any derived classes can be seen as components. Derived classes
    can/should use various world attributes to visualize the simulation. `Matplotlib <https://matplotlib.org/>`_ and
    `irrlicht <http://irrlicht.sourceforge.net/>`_ are two example visualizations that, depending on the simulation
    configuration, users may select. Additional visualizations may be implemented, as well.
    """

    @abstractmethod
    def synchronize(self, time: float):
        pass

    @abstractmethod
    def advance(self, step: float):
        pass

    @abstractmethod
    def is_ok(self) -> bool:
        pass

    @abstractmethod
    def visualize(self, assets, *args, **kwargs):
        """Helper method that visualizes some object(s) in the chosen visualizer

        Different visualizations will visualize the object(s) in different ways. This is an abstract method,
        so it must be implemented.

        Args:
            assets (list): The object(s) to visualize
            *args: Positional arguments that are specific to the underyling visualizer implementation
            **kwargs: Keyworded arguments that are specific to the underlying visualizer implementation
        """
        pass

# -------------------------
# Matplotlib helper classes
# -------------------------


def _transform(entity: np.ndarray, x: float, y: float, yaw: float, alpha: float = 0, x_offset: float = 0, y_offset: float = 0):
    """Helper function to transfrom a numpy entity by the specified values

    There are two transformations that occur. First is considered "static". A simple transformation matrix is constructed where
    the array is rotated by some yaw about the z and a translation from (0,0). A second transformation matrix is then dotted to
    "add" another transform that will account for the previous transformations. This second transformation is used for wheels which
    have the same rotation and position as the chassis, but also are offset and rotated by some value depending on the steering
    angle and the position of the tire relative to the COM.

    Args:
       entity (np.ndarray): The numpy array to transform
       x (float): The static x translation (offset will be applied post rotation)
       y (float): The static y translation (offset will be applied post rotation)
       yaw (float): The static rotation (alpha applied post this rotation)
       alpha (float): The steering angle rotation applied in the second matrix
       x_offset (float): The offset translation in the x direction applied in the second matrix
       y_offset (float): The offset translation in the y direction applied in the second matrix
    """
    T = np.array(
        [[np.cos(yaw), np.sin(yaw), x],
         [-np.sin(yaw), np.cos(yaw), y], [0, 0, 1]]
    )
    T = T @ np.array(
        [
            [np.cos(alpha), np.sin(alpha), x_offset],
            [-np.sin(alpha), np.cos(alpha), y_offset],
            [0, 0, 1],
        ]
    )
    return T.dot(entity)


class _MatplotlibVehicle:
    """Plotter class for a matplotlib vehicle

    Required visualization properties (keys present in the :code:`properties` dictionary):
        "Tire Distance To Front" (float): Offset distance between the COM and the front axle
        "Tire Distance To Rear" (float): Offset distance between the COM and the rear axle
        "Body Distance To Front" (float): Offset distance between the COM and the front bumper
        "Body Distance To Rear" (float): Offset distance between the COM and the rear bumper
        "Body Width" (float): Width of the chassis
        "Tire Width" (float): Tread width of each tire
        "Tire Diameter" (float): Diameter of each tire (in a 2D visualization, will just be length)
        "Track Width" (float): Distance between the wheels on either side of the vehicle


    Args:
        properties (dict): The properties associated with the vehicle representation. See above for required keys.
    """

    def __init__(self, properties: dict):
        body_Lf = properties["Body Distance to Front"]
        body_Lr = properties["Body Distance to Rear"]
        body_width = properties["Body Width"]
        tire_width = properties["Tire Width"]
        tire_diameter = properties["Tire Diameter"]

        self._Lf = properties["Tire Distance to Front"]
        self._Lr = properties["Tire Distance to Rear"]
        self._track_width = properties["Track Width"]
        self._wheelbase = self._Lf + self._Lr

        # Visualization shapes
        self._outline = np.array(
            [
                [-body_Lr, body_Lf, body_Lf, -body_Lr, -body_Lr],
                [
                    body_width / 2,
                    body_width / 2,
                    -body_width / 2,
                    -body_width / 2,
                    body_width / 2,
                ],
                [1, 1, 1, 1, 1],
            ]
        )

        wheel = np.array(
            [
                [
                    tire_diameter,
                    -tire_diameter,
                    -tire_diameter,
                    tire_diameter,
                    tire_diameter,
                ],
                [-tire_width, -tire_width, tire_width, tire_width, -tire_width],
                [1, 1, 1, 1, 1],
            ]
        )

        self._rr_wheel = np.copy(wheel)
        self._rl_wheel = np.copy(wheel)
        self._rl_wheel[1, :] *= -1

        self._fr_wheel = np.copy(wheel)
        self._fl_wheel = np.copy(wheel)
        self._fl_wheel[1, :] *= -1

    def initialize(self, ax, cabcolor: str = "-k", wheelcolor: str = "-k"):
        """Initialize the plotting window for the matplotlib vehicle

        Will save matplotlib returned objects to be used for plotting later.

        Args:
            ax (matplotlib.axes): the axes to plot the vehicle on
            cabcolor (str): The cab color in matplotlib readable format. Defaults to solid black.
            wheelcolor (str): The wheel color in matplotlib readable format. Defaults to solid black.
        """
        cab, = ax.plot(
            np.array(self._outline[0, :]).flatten(),
            np.array(self._outline[1, :]).flatten(),
            cabcolor,
        )
        fr, = ax.plot(
            np.array(self._fr_wheel[0, :]).flatten(),
            np.array(self._fr_wheel[1, :]).flatten(),
            wheelcolor,
        )
        rr, = ax.plot(
            np.array(self._rr_wheel[0, :]).flatten(),
            np.array(self._rr_wheel[1, :]).flatten(),
            wheelcolor,
        )
        fl, = ax.plot(
            np.array(self._fl_wheel[0, :]).flatten(),
            np.array(self._fl_wheel[1, :]).flatten(),
            wheelcolor,
        )
        rl, = ax.plot(
            np.array(self._rl_wheel[0, :]).flatten(),
            np.array(self._rl_wheel[1, :]).flatten(),
            wheelcolor,
        )

        # TODO: Do these variables go out of scope? Are variables copied?
        self._mat_vehicle = (cab, fr, rr, fl, rl)

    def update(self, x: float, y: float, yaw: float, steering: float):
        """Update the state of the matplotlib representation of the vehicle

        Args:
            x (float): x position of the vehicle
            y (float): y position of the vehicle
            yaw (float): yaw rotation of the vehicle about the z (up)
            steering (float): steering angle of the vehicle
        """

        # TODO: I think irrlicht is backwards, so flip plotting so steering and visuals match
        yaw *= -1
        steering *= -1

        # Update the position of each entity
        # All updates are copies of the original array, so each update is basically a fresh update
        fr_wheel = _transform(self._fr_wheel, x, y, yaw, alpha=steering, x_offset=self._Lf, y_offset=-self._track_width)
        fl_wheel = _transform(
            self._fl_wheel,
            x,
            y,
            yaw,
            alpha=steering,
            x_offset=self._Lf,
            y_offset=self._track_width,
        )
        rr_wheel = _transform(
            self._rr_wheel, x, y, yaw, x_offset=-self._Lr, y_offset=-self._track_width
        )
        rl_wheel = _transform(
            self._rl_wheel, x, y, yaw, x_offset=-self._Lr, y_offset=self._track_width
        )
        outline = _transform(self._outline, x, y, yaw)

        (cab, fr, rr, fl, rl) = self._mat_vehicle

        cab.set_ydata(np.array(outline[1, :]).flatten())
        cab.set_xdata(np.array(outline[0, :]).flatten())
        fr.set_ydata(np.array(fr_wheel[1, :]).flatten())
        fr.set_xdata(np.array(fr_wheel[0, :]).flatten())
        rr.set_ydata(np.array(rr_wheel[1, :]).flatten())
        rr.set_xdata(np.array(rr_wheel[0, :]).flatten())
        fl.set_ydata(np.array(fl_wheel[1, :]).flatten())
        fl.set_xdata(np.array(fl_wheel[0, :]).flatten())
        rl.set_ydata(np.array(rl_wheel[1, :]).flatten())
        rl.set_xdata(np.array(rl_wheel[0, :]).flatten())


class _MatplotlibSimpleDashboard:
    """Simple dashboard with selected information that will be displayed within a matplotlib plot window

    TODO: Can we make this a base class and have multiple dashboards (i.e. just make this configurable)
    """

    def initialize(self, ax):
        """Initialize the annotation

        Args:
            ax (matplotlib.axes): The axes to place the annotation on
        """

        bbox_props = dict(boxstyle="round", fc="w", ec="0.5", alpha=0.9)
        self._annotation = ax.annotate(
            "",
            xy=(0.97, 0.7),
            xytext=(0, 10),
            xycoords=("axes fraction", "figure fraction"),
            textcoords="offset points",
            size=10,
            ha="right",
            va="bottom",
            bbox=bbox_props,
        )

    def update(self, vehicle_inputs: 'WAVehicleInputs', time: float, speed: float):
        """Update the dashboard

        Args:
            vehicle_inputs (WAVehicleInputs): The inputs for the vehicle. Visualized in the simple annotation.
            time (float): The current time of the simulation
            speed (float): The instantaneous speed of the vehicle
        """

        text = (
            f"Time :: {time:.2f}\n"
            f"Steering :: {vehicle_inputs.steering:.2f}\n"
            f"Throttle :: {vehicle_inputs.throttle:.2f}\n"
            f"Braking :: {vehicle_inputs.braking:.2f}\n"
            f"Speed :: {speed:.2f}"
        )
        self._annotation.set_text(text)


class _MatplotlibBody:
    """Simple body class for updateable assets

    Args:
        body (WABody): the body to visualize
    """

    def __init__(self, body):
        self._body = body

    def initialize(self, ax):
        """Initialize the body representation

        Args:
            ax (matplotlib.axes): The axes to place the annotation on
        """
        print(self._body.name)
        pass

    def update(self):
        """Update the body"""
        pass


class _MatplotlibBodyList:
    def __init__(self, bodies):
        self.bodies = bodies


class _MatplotlibPlotter(WABase):
    """Base class for matplotlib visualization

    Provides base implementation to be used for all subsequent implementors.

    Args:
        mat_vehicle (_MatplotlibVehicle): The matplotlib vehicle representation
        dashboard (_MatplotlibSimpleDashboard): The dashboard that displays information related to the vehicle state
        record (bool): If set to true, images will be saved under record_filename.
        record_folder (str): The folder to save images to.
        static (bool): if desireable, the matplotlib window can update its axes to "follow" the vehicle. Default is False (follows the vehicle)
        padding (float): Padding between the vehicle's COM and the border of the plot window. Default is 25
        xlim (tuple): x axis limits that are used if static is set to True
        ylim (tuple): y axis limits that are used if static is set to True
    """

    class _Event:
        """Event object that maintains the name of the event and the returned value.

        In matplotlib, callbacks have a name and typically a few class attributes available.
        In this simulator, we only need to pass around the values we want.

        Args:
            name (str): The name of the event
            value (any): The value associated with the event

        Attributes:
            name (str): The name of the event
            value (any): The value associated with the event
        """

        def __init__(self, name: str, value):
            self.name = name
            self.value = value

    def __init__(self, mat_vehicle: _MatplotlibVehicle, dashboard: _MatplotlibSimpleDashboard, opponent_mat_vehicles: list, record: bool, record_folder: str, static: bool = False, padding: float = 25, xlim: tuple = (-50, 50), ylim: tuple = (50, -50)):
        self._mat_vehicle = mat_vehicle
        self._dashboard = dashboard
        self._opponent_mat_vehicles = opponent_mat_vehicles

        self._record = record
        self._record_folder = record_folder

        if self._record:
            exit(0)

        self._save_number = 0

        self._vehicle_inputs = None
        self._state = (0, 0, 0, 0)
        self._time = 0

        self._static = static
        self._padding = padding
        self._xlim = xlim
        self._ylim = ylim

        # Asynchronous Matplotlib events
        self._events = dict()

    def set_vehicle(self, vehicle: 'WAVehicle'):
        """Set the vehicle object

        Args:
            vehicle (WAVehicle): The vehicle
        """
        self._vehicle = vehicle

    def set_vehicle_inputs(self, vehicle_inputs: 'WAVehicleInputs'):
        """Set the vehicle inputs object

        Args:
            vehicle_inputs (WAVehicleInputs): The vehicle inputs
        """
        self._vehicle_inputs = vehicle_inputs

    def set_opponents(self, opponents: list):
        """Set the opponents list

        Args:
            opponents (list): The list of opponents
        """
        self._opponents = opponents
        self._opponent_states = [[(0, 0, 0, 0), opponent._vehicle_inputs] for opponent in opponents]

    def synchronize(self, time: float):
        # Save values for later
        self._time = time

    def advance(self, step: float):
        # Update the state
        pos = self._vehicle.get_pos()
        yaw = self._vehicle.get_rot().to_euler_yaw()
        v = self._vehicle.get_pos_dt().length
        self._state = (pos.x, pos.y, yaw, v)

        for i, opponent in enumerate(self._opponents):
            pos = opponent.get_pos()
            yaw = opponent.get_rot().to_euler_yaw()
            v = opponent.get_pos_dt().length
            self._opponent_states[i][0] = (pos.x, pos.y, yaw, v)

    def _savefig(self):
        """Save a figure from matplotlib"""

        if self._record:
            if self._save_number == 0:
                import os
                if not os.path.exists(self._record_folder):
                    os.makedirs(self._record_folder)
                del os

            self._fig.savefig(self._record_folder + f'{self._save_number}.png', dpi=200)

            self._save_number += 1

    @abstractmethod
    def register_key_press_event(self, callback):
        """Register callback with a key press event in the matplotlib window

        # matplotlib.backend_bases.KeyEvent>`_ provide an event
        `Matplotlib KeyEvents <https://matplotlib.org/stable/api/backend_bases_api.html
        object with various attributes. In the case of controllers, we only really need the keypressed and not necessarily the location of anything.

        .. todo::
            Add support for more general callbacks (i.e. mouse)

        Args:
            callback (method(int)): A callable method that takes in an int (represents the key pressed)
        """
        pass

    @abstractmethod
    def plot(self, *args, **kwargs):
        """Plot additional information within the matplotlib window

        Args:
            *args: positional arguments that are passed to the underyling plotter (and subsequently to matplotlib)
            **kwargs: keyworded arguments that are passed to the underyling plotter (and subsequently to matplotlib)
        """
        pass

    def _initialize_plot(self):
        """Helper method that initializes a matplotlib plot and passes the axes to the vehicle and dashboard"""

        # Initial plotting setup
        self._fig, self._ax = plt.subplots(figsize=(8, 8))

        # Plot styling
        self._ax.set_xlim(*self._xlim)
        self._ax.set_ylim(*self._ylim)

        self._ax.grid(True)
        self._ax.set_aspect("equal", adjustable="box")

        # Initialize the plotters
        self._mat_vehicle.initialize(self._ax)
        self._dashboard.initialize(self._ax)
        for o in self._opponent_mat_vehicles:
            o.initialize(self._ax)

    def _update_axes(self, x, y):
        """Helper method to update the matplotlib axis plot.

        Moving vehicles commonly want to be "followed". This method will place the vehicle at the center of the window and adjust
        the axes accordingly. The padding specified at initialization will be added.

        Args:
            x (float): x position of the vehicle
            y (float): y position of the vehicle
        """
        self._ax.set_xlim(x - self._padding, x + self._padding)
        self._ax.set_ylim(y + self._padding, y - self._padding)


class _MatplotlibSinglePlotter(_MatplotlibPlotter):
    """Single process plotter for sequential visualization. Can be used in jupyter.

    A single threaded plotter basically will update it's state `in order` with other components. This can be thought of
    as the typical use case for matplotlib where :code:`plt.show()`/:code:`plt.pause()` comes directly after `plt.plot`.

    For jupyter, we need to use IPython to basically create an image and display that image in the jupyter browser. This really sucks
    cause it's super slow. TODO: Is there a faster way to do that?

    Args:
        mat_vehicle (_MatplotlibVehicle): The matplotlib vehicle representation
        dashboard (_MatplotlibSimpleDashboard): The dashboard that displays information related to the vehicle state
        is_jupyter (bool): should jupyter visualization be used (i.e. IPython display)?
        record (bool, optional): If set to true, images will be saved under record_filename. Defaults to False (doesn't save images).
        record_folder (str, optional): The folder to save images to. Defaults to "OUTPUT/".
        **kwargs: keyworded arguments used for the base _MatplotlibPlotter class
    """

    def __init__(self, mat_vehicle, dashboard, opponent_mat_vehicles, is_jupyter: bool = False, record: bool = False, record_folder: str = "OUTPUT/", **kwargs):
        super().__init__(mat_vehicle, dashboard, opponent_mat_vehicles, record, record_folder, **kwargs)

        self._initialize_plot()

        self._is_jupyter = is_jupyter

    def advance(self, step):
        super().advance(step)

        # Extract values
        x, y, yaw, v = self._state

        # Update the simulation elements
        self._mat_vehicle.update(x, y, yaw, self._vehicle_inputs.steering)
        self._dashboard.update(self._vehicle_inputs, self._time, v)

        # Update the axes to "follow" the vehicle
        if not self._static:
            self._update_axes(x, y)

        for i, (state, vehicle_inputs) in enumerate(self._opponent_states):
            x, y, yaw, v = state
            self._opponent_mat_vehicles[i].update(x, y, yaw, vehicle_inputs.steering)

        # Update the plot
        if self._is_jupyter:
            display(self._fig)
            clear_output(wait=True)
        else:
            plt.pause(1e-9)

        # Save (if desired)
        self._savefig()

    def is_ok(self):
        return plt.fignum_exists(self._fig.number)

    def register_key_press_event(self, callback):
        self._fig.canvas.mpl_connect('key_press_event', self._key_press)
        self._events['key_press_event'] = callback

    def plot(self, *args, **kwargs):
        if isinstance(args[0], Patch):
            self._ax.add_patch(args[0])
        elif isinstance(args[0], collections.Collection):
            self._ax.add_collection(args[0])
        else:
            self._ax.plot(*args, **kwargs)

    def _key_press(self, event):
        """Key press callback. Passes the event key (i.e. 'up' or 'a') to the callback registered through :meth:`~register_key_press_event`.

        Args:
            event (matplotlib.event): event with a name and key attribute
        """
        if event.name in self._events:
            # Call that event with the event key (is a string like 'up' or 'a')
            self._events[event.name](event.key)


class _MatplotlibMultiPlotter(_MatplotlibPlotter):
    """Multi process plotter for sequential visualization. Significantly faster than the sequential :class:`~_MatplotlibSinglePlotter`.

    A multi threaded plotter will basically update it's state at a fixed rate. All visualizations occur in a separate thread, so
    a synchronized queue is used to pass information between threads. At the fixed rate, on an update, that information is grabbed from the front
    of the queue. Stale information is always ignored so there is only ever 1 or 0 packets in the queue. A `asynchronous` constructor parameter
    provides the configurability to either wait for state information to be used or continue and just throw out old state data. The
    latter results in the simulation progressing significantly faster than the plotter, so it's not recommended.

    Two communication methods are used. One is a queue and is used strictly for state information and the other is a Pipe and used for simulation
    communication related items, such as keyboard callback registration.

    Args:
        mat_vehicle (_MatplotlibVehicle): The matplotlib vehicle representation
        dashboard (_MatplotlibSimpleDashboard): The dashboard that displays information related to the vehicle state
        asynchronous (bool, optional): If true, the plotter will continue without waiting for the passed state information to be used. Defaults to synchronous.
        record (bool, optional): If set to true, images will be saved under record_filename. Defaults to False (doesn't save images).
        record_folder (str, optional): The folder to save images to. Defaults to "OUTPUT/".
        **kwargs: keyworded arguments used for the base _MatplotlibPlotter class
    """

    class _SimulationState:
        """Simulation state helper class that just holds various attributes for plotting

        Pickable to pass between threads.

        Args:
            state (tuple): holds (x,y,yaw,steering)
            vehicle_inputs (WAVehicleInputs): the inputs to the vehicle
            time (float): the current time of the simulation

        Attributes:
            state (tuple): holds (x,y,yaw,steering)
            vehicle_inputs (WAVehicleInputs): the inputs to the vehicle
            time (float): the current time of the simulation
        """

        def __init__(self, state: tuple, vehicle_inputs: 'WAVehicleInputs', time: float):
            self.state = state
            self.vehicle_inputs = vehicle_inputs
            self.time = time

        def __call__(self):
            return self.state, self.vehicle_inputs, self.time

    class _PlotCall:
        """Pickable plot call that is passed to the plotter thread.

        Args:
            *args: Positional arguments used in matplotlib.plot
            **kwargs: Keyworded arguments used in matplotlib.plot

        Attributes:
            args: Positional arguments used in matplotlib.plot
            kwargs: Keyworded arguments used in matplotlib.plot
        """

        def __init__(self, *args, **kwargs):
            self.args = args
            self.kwargs = kwargs

    def __init__(self, mat_vehicle: '_MatplotlibVehicle', dashboard: '_MatplotlibSimpleDashboard', opponent_mat_vehicles: list, asynchronous: bool = False, record: bool = False, record_folder: str = "OUTPUT/", **kwargs):
        super().__init__(mat_vehicle, dashboard, opponent_mat_vehicles, record, record_folder, **kwargs)

        self._asynchronous = asynchronous

        self._events['close_event'] = self._close

        # Initialize multiprocessing objects
        if plt.get_backend() == "MacOSX":
            set_start_method("spawn", force=True)

        # Communication pipelines between processes
        self._event_sender, self._event_receiver = Pipe()
        self._queue = Queue(1)

        # Barrier that will wait for initialization to occur
        self._barrier = Barrier(2)

        self._p = Process(target=self.run)
        self._p.start()

        # Wait for the matplotlib window to initialize
        self._barrier.wait()

    def advance(self, step):
        super().advance(step)

        if self._event_receiver.poll():
            event = self._event_receiver.recv()
            if event.name in self._events:
                self._events[event.name](event.value)

        # Send the state to the other process to visualize
        def send(state):
            if not self._queue._closed:
                if self._asynchronous and self._queue.empty():
                    self._queue.put_nowait(state)
                else:
                    self._queue.put(state)

        states = [self._SimulationState(self._state, self._vehicle_inputs, self._time)]
        for state, vehicle_inputs in self._opponent_states:
            states.append(self._SimulationState(state, vehicle_inputs, self._time))
        send(states)

    def run(self):
        """Multiprocess starter method. Will initialize matplotlib and setup FuncAnimation"""

        # Initialize matplotlib
        self._initialize_plot()
        self._fig.canvas.mpl_connect('close_event', self._handle_close)

        # Initialize animation
        anim = FuncAnimation(self._fig, self._update, interval=10)

        # Synchronize with main process
        self._barrier.wait()

        plt.show()

        # Flush the queue
        if not self._queue.empty():
            self._queue.get_nowait()

    def _update(self, i: int):
        """Called at a specific interval by FuncAnimation. Will update matplotlib window

        Args:
            i (int): window count
        """
        try:
            state = self._queue.get(timeout=10 if i <= 2 else 1)
        except:
            plt.close(self._fig)
            return

        if isinstance(state, self._SimulationState):
            (x, y, yaw, v), vehicle_inputs, time = state()

            self._mat_vehicle.update(x, y, yaw, vehicle_inputs.steering)
            self._dashboard.update(vehicle_inputs, time, v)

            if not self._static:
                self._update_axes(x, y)
        elif isinstance(state, list):
            this_state = state[0]
            (x, y, yaw, v), vehicle_inputs, time = this_state()

            self._mat_vehicle.update(x, y, yaw, vehicle_inputs.steering)
            self._dashboard.update(vehicle_inputs, time, v)

            if not self._static:
                self._update_axes(x, y)

            for i, opponent in enumerate(state[1:]):
                (x, y, yaw, v), vehicle_inputs, time = opponent()

                self._opponent_mat_vehicles[i].update(x, y, yaw, vehicle_inputs.steering)

            # Save (if desired)
            self._savefig()

        elif isinstance(state, str):
            if state == 'key_press_event':
                self._fig.canvas.mpl_connect('key_press_event', self._key_press)
        elif isinstance(state, self._PlotCall):
            if isinstance(state.args[0], Patch):
                self._ax.add_patch(state.args[0])
            elif isinstance(state.args[0], collections.Collection):
                self._ax.add_collection(state.args[0])
            elif isinstance(state.args[0], _MatplotlibBodyList):
                for body in state.args[0].bodies:
                    body.initialize(self._ax)
            else:

                self._ax.plot(*state.args, **state.kwargs)

    def is_ok(self) -> bool:
        return self._p.is_alive()

    def plot(self, *args, **kwargs):
        self._queue.put(self._PlotCall(*args, **kwargs))

    def register_key_press_event(self, callback):
        self._queue.put('key_press_event')
        self._events['key_press_event'] = callback

    def _handle_close(self, event):
        """Callback from matplotlib when a plot is closed"""
        self._event_sender.send(self._Event(event.name, 'close_event'))

    def _key_press(self, event):
        """Key press callback. Passes the event key (i.e. 'up' or 'a') to the callback registered through :meth:`~register_key_press_event`.

        Args:
            event (matplotlib.event): event with a name and key attribute
        """
        self._event_sender.send(self._Event(event.name, event.key))

    def _close(self, event):
        """Close callback"""
        self._queue.close()
        self._event_receiver.close()
        self._event_sender.close()
        self._p.join()

    def __del__(self):
        """Destructor"""
        if hasattr(self, "p") and self._p.is_alive():
            self._p.join()


class WAMatplotlibVisualization(WAVisualization):
    """Matplotlib visualizer of the simulator world and the vehicle

    Plotter requires a vehicle and should be used for vehicle simulations where a 2d world representation is adequate. Furthermore,
    for vehicle models that aren't implemented through Chrono, this is the only implemented visualization for users.

    There are three types of matplotlib plotters currently implemented: "single", "multi" and "jupyter". Each has a different purpose with the same
    goal (to visualize the simulation).

    * single:

      * Uses a single thread where visualization updates are done sequentially with other simulation components.

      * *Worse performance than multi*.

    * multi:

      * Uses two threads where visualization updates are placed on a separate threads and updates are asynchronous. State information of the environment and vehicle are passed to the plotter on each update, but updates of the visuals are done at a fixed rate(10 [Hz]).

      * *Better performance than single*.

    * jupyter:

      * Supports visualization of a matplotlib window in a `jupyter notebook <https://jupyter.org/>`_.

      * *Performance is very poor*. *Do not use outside of jupyter*.

    Args:
        system (WASystem): system used to grab certain parameters of the simulation
        vehicle (WAVehicle): vehicle to render in the matplotlib plot window
        vehicle_inputs (WAVehicleInputs): the vehicle inputs
        environment (WAEnvironment, optional): An environment with various world assets to visualize. Defaults to None (doesn't visualize anything).
        plotter_type (str, optional): Type of plotter. "single" for single threaded, "multi" for multi threaded (fastest), "jupyter" if jupyter is used. Defaults to "single".
        opponents (list, optional): List of other :class:`~WAVehicle`'s that represent opponents. Used for rendering.
        record (bool, optional): If set to true, images will be saved under record_filename. Defaults to False (doesn't save images).
        record_folder (str, optional): The folder to save images to. Defaults to "OUTPUT/".
        **kwargs: Additional arguments that are based to the underlying plotter implementation.

    Raises:
        ValueError: plotter_type isn't recognized
    """

    def __init__(self, system: 'WASystem', vehicle: 'WAVehicle', vehicle_inputs: 'WAVehicleInputs', environment: 'WAEnvironment' = None,  plotter_type: str = "single", opponents: list = [], record: bool = False, record_folder: str = "OUTPUT/", **kwargs):
        self._system = system
        self._vehicle = vehicle
        self._vehicle_inputs = vehicle_inputs

        self._render_steps = int(np.ceil(system.render_step_size / system.step_size))  # noqa

        # Create the vehicle and dashboard
        mat_vehicle = _MatplotlibVehicle(self._vehicle.get_visual_properties())
        dashboard = _MatplotlibSimpleDashboard()

        opponent_mat_vehicles = []
        for opponent in opponents:
            opponent_mat_vehicles.append(_MatplotlibVehicle(opponent.get_visual_properties()))

        supported_plotters = ['multi', 'single', 'jupyter']

        # Create the underlying plotter
        if plotter_type == "multi":
            self._plotter = _MatplotlibMultiPlotter(
                mat_vehicle, dashboard, opponent_mat_vehicles, record=record, record_folder=record_folder, **kwargs)
        elif plotter_type == "single" or plotter_type == "jupyter":
            is_jupyter = plotter_type == 'jupyter'
            self._plotter = _MatplotlibSinglePlotter(
                mat_vehicle, dashboard, opponent_mat_vehicles, is_jupyter=is_jupyter, record=record, record_folder=record_folder, **kwargs)
        else:
            raise ValueError(
                f"Unknown plotter type: {plotter_type}. Must be one of the following: {', '.join(supported_plotters)}")

        self._plotter.set_vehicle(vehicle)
        self._plotter.set_vehicle_inputs(vehicle_inputs)
        self._plotter.set_opponents(opponents)

        if environment is not None:
            self.visualize(environment.get_assets())

    def synchronize(self, time: float):
        self._plotter.synchronize(time)

    def advance(self, step: float):
        if self._system.step_number % self._render_steps == 0:
            # Will only call advance at the rate specified by render_step_size in the passed system
            self._plotter.advance(step)

    def is_ok(self) -> bool:
        return self._plotter.is_ok()

    def register_key_press_event(self, callback):
        """Register a key press callback with the matplotlib visualization

        Args:
            callback(function): the callback to invoke when a key is pressed
        """
        self._plotter.register_key_press_event(callback)

    def plot(self, *args, **kwargs):
        """Pass plot info to the underyling plotter

        In some cases, additional information will want to be plotted in the plotter window. For example, in a vehicle simulation where
        a controller is attempting to follow a path, that path may want to be visualized in the matplotlib window. This method faciliates
        that functionality.

        See the `matplotlib docs <https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.plot.html>`_ to see additional *args and **kwargs arguments.

        Args:
            *args: positional arguments that are passed to the underyling plotter (and subsequently to matplotlib)
            **kwargs: keyworded arguments that are passed to the underyling plotter (and subsequently to matplotlib)
        """
        self._plotter.plot(*args, **kwargs)

    def visualize(self, assets, *args, **kwargs):
        if not isinstance(assets, list):
            assets = [assets]

        patches = []
        bodies = []

        for i, asset in enumerate(assets):
            if isinstance(asset, WAPath):
                points = asset.get_points()
                self._plotter.plot(points[:, 0], points[:, 1], *args, **asset.get_vis_properties(), **kwargs)
            elif isinstance(asset, WATrack):
                self.visualize(asset.left, *args, **kwargs)
                self.visualize(asset.right, *args, **kwargs)
                self.visualize(asset.center, *args, **kwargs)
            elif isinstance(asset, WABody):
                if not hasattr(asset, 'size') or not hasattr(asset, 'position'):
                    raise AttributeError("Body must have 'size', and 'position' fields")

                if hasattr(asset, 'updates') and asset.updates:
                    # bodies.append(_MatplotlibBody(asset))
                    print(
                        "WARNING: updateable WABody's are currently not supported for the WAMatplotlibVisualization. This will come in a future release.")
                    continue

                position = asset.position
                yaw = 0 if not hasattr(asset, 'yaw') else asset.yaw
                size = asset.size / 2
                color = WAVector([1, 0, 0]) if not hasattr(asset, 'color') else asset.color

                # Can't really see white
                edgecolor = color
                if color == WAVector([1, 1, 1]):
                    edgecolor = WAVector([0, 0, 0])

                outline = np.array([[-size.y, size.y, size.y, -size.y, -size.y],
                                    [size.x, size.x, -size.x, -size.x, size.x],
                                    [1, 1, 1, 1, 1]])

                outline = _transform(outline, position.x, position.y, yaw)

                patches.append(Polygon(outline[:2, :].T, facecolor=color, edgecolor=edgecolor, alpha=0.4))

        if len(patches):
            self._plotter.plot(collections.PatchCollection(patches, match_original=True))
        if len(bodies):
            self._plotter.plot(_MatplotlibBodyList(bodies))



"""
Wisconsin Autonomous - https://www.wisconsinautonomous.org

Copyright (c) 2021 wisconsinautonomous.org
All rights reserved.

Use of this source code is governed by a BSD-style license that can be found
in the LICENSE file at the top level of the repo
"""


# WA Simulator
from wa_simulator.path import WAPath, create_path_from_json
from wa_simulator.core import WAVector
from wa_simulator.utils import _load_json, _check_field, get_wa_data_file

# Other imports
import numpy as np
import matplotlib.pyplot as plt


def create_track_from_json(filename: str, environment: 'WAEnvironment' = None) -> 'WATrack':
    """Creates a WATrack object from a json specification file

    json file options:

    * Center Input File (``str``, required): A json file describing the centerline. Loaded using :meth:`~create_path_from_json`

    * Width (``float``, required): The constant width between the left and right boundaries of the track.

    * Visualization (``dict``, optional): Additional visualization properties.

      * Center/Right/Left (``dict``, optional): The each paths visualization properties

        * Color (``list``, optional): Visualization color.

        * Object (``dict``, optional): An object that is placed along the path. Only parsed if ``environment`` is set.

          * Size (``list``, optional): Size of the objects.

          * Color (``list``, optional): Color of the objects.

          * Color #1 (``list``, optional): Color of an alternating set of objects. Must come with Color #2 and without Color.

          * Color #2 (``list``, optional): Color of an alternating set of objects. Must come with Color #1 and without Color.

          * Mode (``str``, optional): The mode for the object placement along the path. Options include 'Solid', 'Dashed' (3[m] separation) and 'Spread' (6[m] separation).

    .. todo::

        Add a variable width loader

    Args:
        filename (str): The json specification that describes the track 
        environment (WAEnvironment, optional): Adds objects to the environment, if present. Defaults to None (doesn't parse objects).

    Returns:
        WATrack: The created track
    """

    j = _load_json(filename)

    # Validate the json file
    _check_field(j, 'Type', value='Track')
    _check_field(j, 'Template', allowed_values='Constant Width Track')
    _check_field(j, 'Center Input File', field_type=str)
    _check_field(j, 'Width', field_type=float)
    _check_field(j, 'Visualization', field_type=dict, optional=True)

    # Create the centerline path
    center_file = get_wa_data_file(j['Center Input File'])
    center = create_path_from_json(center_file)

    width = j['Width']

    # Create the track
    track = create_constant_width_track(center, width)

    # Load the visualization, if present
    if 'Visualization' in j:
        v = j['Visualization']
        _check_field(v, 'Center', field_type=dict, optional=True)
        _check_field(v, 'Left', field_type=dict, optional=True)
        _check_field(v, 'Right', field_type=dict, optional=True)

        def _load_vis(path, path_name):
            if path_name in v:
                p = v[path_name]

                _check_field(p, 'Color', field_type=list, optional=True)
                _check_field(p, 'Object', field_type=dict, optional=True)

                if 'Color' in p:
                    path.get_vis_properties()['color'] = WAVector(p['Color'])

                if 'Object' in p and environment is not None:
                    o = p['Object']

                    _check_field(o, 'Size', field_type=list)
                    _check_field(o, 'Color', field_type=list, optional=True)
                    _check_field(o, 'Mode', field_type=str, optional=True)

                    kwargs = {}
                    kwargs['size'] = WAVector(o['Size'])

                    if 'Color' in o:
                        kwargs['color'] = WAVector(o['Color'])

                        if 'Color #1' in o or 'Color #2' in o:
                            raise ValueError("'Color' cannot be used with 'Color #1' or 'Color #2'")
                    elif 'Color #1' in o and 'Color #2' in o:
                        kwargs['color1'] = WAVector(o['Color #1'])
                        kwargs['color2'] = WAVector(o['Color #2'])
                    elif 'Color #1' in o or 'Color #2' in o:
                        raise ValueError("'Color #1' and 'Color #2' must be used together.")

                    s = path.calc_length_cummulative()[-1]
                    size = WAVector(o['Size'])
                    if 'Mode' in o:
                        m = o['Mode']

                        if m == 'Continuous':
                            n = s / size.y
                        elif m == 'Dashed':
                            n = s / 3  # Spaced as dashed center lines are (3[m] apart)
                        elif m == 'Spread':
                            n = s / 6
                        else:
                            raise ValueError(f"'{m}' is not a supported road marker type")
                    else:
                        # Defaults to continuous
                        n = s / size.y

                    points = path.get_points()
                    d_points = path.get_points(der=1)

                    s = path.calc_length_cummulative()[-1]
                    size = WAVector(o['Size'])

                    l = len(points)
                    for e, i in enumerate(range(0, l, 1 if l < n else int(l / n))):
                        p = points[i]
                        dp = d_points[i]

                        kwargs['position'] = WAVector(p)
                        kwargs['yaw'] = -np.arctan2(dp[1], dp[0])

                        if 'color1' in kwargs:
                            if e % 2 == 0:
                                kwargs['color'] = kwargs['color1']
                            else:
                                kwargs['color'] = kwargs['color2']

                        environment.create_body(**kwargs)

        _load_vis(track.center, 'Center')
        _load_vis(track.right, 'Right')
        _load_vis(track.left, 'Left')

    return track


def create_constant_width_track(center: WAPath, width: float) -> 'WATrack':
    """Generates a WAConstantWidthTrack given a centerline and a constant width. Simply "walks" along path and takes the normal from the center at a distance equal to width/2

    Args:
        center (WAPath): The centerline of the track
        width (float): The constant distance between the left and right boundaries(distance between centerline and a boundary is width/2)

    Returns:
        WATrack: The created track
    """
    if center._d_points is None:
        raise ValueError(
            'create_constant_width_track: derivative of the centerline has not been initialized')

    left, right = [], []

    points = center._points
    d_points = center._d_points

    for i in range(len(points)-1):
        ix, iy, iz = points[i]
        d_point = d_points[i]

        l = np.linalg.norm(d_point)
        dx = d_point[0] * width / (2 * l)
        dy = d_point[1] * width / (2 * l)

        left.append([ix - dy, iy + dx, iz])
        right.append([ix + dy, iy - dx, iz])

    def close_path_if_necessary(points):
        if center.is_closed() and not np.array_equal(points[0], points[-1]):
            return np.vstack((points, points[0]))
        return np.array(points)

    left = close_path_if_necessary(left)
    right = close_path_if_necessary(right)

    # No interpolation to maintain normals
    left_path = type(center)(left, **center.get_parameters())
    right_path = type(center)(right, **center.get_parameters())

    return WATrack(center, left_path, right_path, width=width)


class WATrack:
    """Base Track object. Basically holds three WAPaths: centerline and two boundaries. This class provides convenience functions so that it is easier to write various track related code

    Does *not* inherit from WABase. This is a static component, so it does not need to update.

    Args:
        center(WAPath): The centerline of the track.
        left(WAPath): The "left" side boundary. Side is determined by orientation from the centerline by index.
        right(WAPath): The "right" side boundary. Side is determined by orientation from the centerline by index.
        kwargs: Extra parameters that are used for the track

    Attributes:
        center(WAPath): The centerline of the track.
        left(WAPath): The "left" side boundary. Side is determined by orientation from the centerline by index.
        right(WAPath): The "right" side boundary. Side is determined by orientation from the centerline by index.
    """

    def __init__(self, center: WAPath, left: WAPath, right: WAPath, **kwargs):
        self.center = center
        self.left = left
        self.right = right

        # Will expand out the extras dictionary into class variables
        self.__dict__.update(kwargs)

    def inside_boundaries(self, point: WAVector) -> bool:
        """Check whether the passed point is within the track boundaries

        Implementation is explained `here <https://stackoverflow.com/a/33155594>`_.

        Args:
            point(WAVector): point to check whether it's inside the track boundaries

        Returns:
            bool: is the point inside the boundaries?
        """
        if not isinstance(point, WAVector):
            raise TypeError(
                f'WATrack.inside_boundary: Expects a WAVector, not a {type(point)}.')

        closest_point, idx = self.center.calc_closest_point(point, True)
        A, B, C = point, WAVector(self.left.get_points()[idx]), WAVector(self.right.get_points()[idx])  # noqa
        a, b, c = (B-C).length, (C-A).length, (A-B).length
        return a**2 + b**2 >= c**2 and a**2 + c**2 >= b**2

    def plot(self, show=True, center_args: dict = {}, left_args: dict = {}, right_args: dict = {}):
        """Plot the path. Most likely plotter is matplotlib, but technically anything can be used.

        Args:
            show(bool): Immediately show the plot? Defaults to True.
            center_args(dict): Keyworded parameters passed to the center plot method
            left_args(dict): Keyworded parameters passed to the left plot method
            right_args(dict): Keyworded parameters passed to the right plot method
        """
        self.center.plot(**center_args, show=False)
        self.left.plot(**left_args, show=False)
        self.right.plot(**right_args, show=False)

        if show:
            plt.show()



"""
Wisconsin Autonomous - https://www.wisconsinautonomous.org

Copyright (c) 2021 wisconsinautonomous.org
All rights reserved.

Use of this source code is governed by a BSD-style license that can be found
in the LICENSE file at the top level of the repo
"""

from abc import ABC, abstractmethod  # Abstract Base Class

# WA Simulator
from wa_simulator.base import WABase


class WASimulationManager(WABase):
    """A manager for a simulation. Advances and synchronizes simulation modules.

    The simulation manager is used primarily for cleaning up demo and runner files.
    Typically, users do not want to have all update methods(synchronize / advance) in their actual demo file, as that would convalute the
    demo and make things more complicated. This class aims to simplify these scenarios
    buy wrapping update related method calls not necessarily relevant to the user.

    This class constructor (the :code:`__init__` method) has two actual arguments: :code:`*args` and :code:`**kwargs`.
    These two arguments have special attributes not relevant in other languages, so for more
    information on those, please see `this reference <https://www.digitalocean.com/community/tutorials/how-to-use-args-and-kwargs-in-python-3>`_.

    Although not techincally an error, this class should not be inherited from.

    Args:
        system (WASystem): the system that has meta properties such as time and render step size
        *args: Positional arguments made up of WABase classes. An exception will be thrown if the components are not WABase level classes
        record (boolean): Log simulation data as a csv?
        output_filename (str): Only used if record is true. If record is true and this is empty, an exception will be raised.

    Raises:
        ValueError: If any one of the positional arguments does not derive from WABase (`see classes that do <google.com>`_).
        ValueError: If record is true but output_filename is empty.

    .. todo::
        - Add link to all WABase classes

    Examples:

    .. highlight:: python
    .. code-block:: python

        import wa_simulator as wa

        ... # Initialization of subcomponents

        # Ex 1.
        manager = wa.WASimulationManager(
            vehicle, visualization, system, controller, environment)

        # Ex 2.
        manager = wa.WASimulationManager(sensor_manager, system, environment)

        # Ex 3.
        # WARNING: Recording is not currently supported
        manager = wa.WASimulationManager(vehicle, visualization, system, controller,
                                     environment, sensor_manager, record=True, output_filename="example.csv")
    """

    def __init__(self, system: 'WASystem', *args, record: bool = False, output_filename: str = ""):
        self._system = system

        self._components = []
        for i, comp in enumerate(args):
            if comp is None:
                continue

            # Check to make sure each element derives from WABase
            if not isinstance(comp, WABase):
                raise TypeError(
                    f'Argument in position {i} is of type {type(comp)}, not {type(WABase)}.')
            elif isinstance(comp, WASimulationManager):
                raise TypeError(
                    f'Argument in position {i} is of type {type(WASimulationManager)}. This is not allowed')

            self._components.append(comp)

        self.set_record(record, output_filename)

    def synchronize(self, time: float):
        for comp in self._components:
            comp.synchronize(time)

    def advance(self, step: float):
        self._system.advance()

        for comp in self._components:
            comp.advance(step)

    def is_ok(self) -> bool:
        for comp in self._components:
            if not comp.is_ok():
                return False
        return self._system.is_ok()

    def run(self):
        """Helper method that runs a simulation loop. 

        It's recommended to just use :meth:`~synchronize` and :meth:`~advance` to remain explicit in intent,
        but this method can also be used to update states of components. Basically, this method will just
        call the :meth:`~synchronize` and :meth:`~advance` functions in a :code:`while` loop until any components
        fail.
        """
        step_size = self._system.step_size
        while self.is_ok():
            time = self._system.time

            self.synchronize(time)
            self.advance(step_size)

    def record(self):
        """Perform a record step for each component active in the simulation.

        .. warning::
            Currently not implemented
        """
        pass

    def set_record(self, record: bool = True, output_filename: str = ""):
        """Specify whether the simulation should be recorded.

        Args:
            record (bool): record the simulation?
            output_filename (str): the csv filename to save data to. If record is true and this is empty, an exception will be raised
        """
        self._record = record
        self._output_filename = output_filename

        if record and len(output_filename) == 0:
            raise AttributeError(
                "record has been set to 'True', but output_filename is empty.")



"""
Wisconsin Autonomous - https://www.wisconsinautonomous.org

Copyright (c) 2021 wisconsinautonomous.org
All rights reserved.

Use of this source code is governed by a BSD-style license that can be found
in the LICENSE file at the top level of the repo
"""


from abc import abstractmethod  # Abstract Base Class

# WA Simulator
from wa_simulator.core import WAVector
from wa_simulator.utils import _load_json, _check_field, get_wa_data_file

# Other imports
import numpy as np
import matplotlib.pyplot as plt
import warnings
from scipy.interpolate import splprep, splev
from scipy.spatial.distance import cdist


def create_path_from_json(filename: str) -> 'WAPath':
    """Creates a WAPath object from json

    json file options:

    * Waypoints Input File (str, required): A csv file describing the path waypoints. Loaded using :meth:`~load_waypoints_from_csv`.

    * Additional keyworded arguments necessary for the path template

    Args:
        filename (str): The json specification that describes the path
    """

    j = _load_json(filename)

    # Validate the json file
    _check_field(j, 'Type', value='Path')
    _check_field(j, 'Template', allowed_values=['WASplinePath'])
    _check_field(j, 'Waypoints Input File', field_type=str)

    # Grab the waypoints
    waypoints_file = get_wa_data_file(j['Waypoints Input File'])
    waypoints = load_waypoints_from_csv(waypoints_file, delimiter=",")

    excluded_keys = ['Type', 'Template', 'Waypoints Input File']
    kwargs = {x: j[x] for x in j if x not in excluded_keys}

    # Create the path
    path = eval(j['Template'])(waypoints, **kwargs)

    return path


def load_waypoints_from_csv(filename: str, **kwargs) -> np.ndarray:
    r"""Get data points from a csv file.

    Should be structured as "x,y,z\\nx,y,z...". See `NumPy.loadtxt < https: // numpy.org/doc/stable/reference/generated/numpy.loadtxt.html >`_
    for more info on arguments.

    Args:
        filename(str): file to open and read data from

    Returns:
        np.ndarray: an n x m array with each data point in each row
    """
    return np.loadtxt(filename, **kwargs)


def calc_path_length_cummulative(x, y) -> np.ndarray:
    """Get the cummulative distance along a path provided the given x and y position values

    Args:
        x(np.ndarray): x coordinates
        y(np.ndarray): y coordinates

    Returns:
        np.ndarray: the cummulative distance along the path
    """
    return np.cumsum(np.linalg.norm(np.diff(np.column_stack((x, y)), axis=0), axis=1))


def calc_path_curvature(dx, dy, ddx, ddy) -> np.ndarray:
    """Calculate the curvature of a path at each point

    Args:
        dx(np.ndarray): first x derivative
        dy(np.ndarray): first y derivative
        ddx(np.ndarray): second x derivative
        ddy(np.ndarray): second y derivative

    Returns:
        np.ndarray: the curvature at each point
    """
    return (dx * ddy - dy * ddx) / (dx ** 2 + dy ** 2) ** (3 / 2)


class WAPath:
    """Base Path object. To be used to generate paths or trajectories for path planning and / or path following

    All path objects * should * be implemented in a 3D coordinate space! This means, waypoints should be a list or np.ndarray of
    lists or np.ndarrays of size 3!

    Example:

    .. highlight:: python
    .. code-block:: python

        from wa_simulator.path import WAPath, load_waypoints_from_csv

        # Simple 2D Path
        waypoints = [
            [1, 2, 0],
            [2, 2, 0],
            [5, 5, 0],
        ]
        # Not actually allowed since WAPath is abstract (has abstract methods)
        path_2D = WAPath(waypoints)

        # Simple 3D Path
        waypoints = [
            [1, 2, 1],
            [2, 2, 2],
            [5, 5, 1],
        ]
        # Not actually allowed since WAPath is abstract (has abstract methods)
        path_3D = WAPath(waypoints)

        # JSON loaded path
        waypoints = load_waypoints_from_csv("path.csv")
        # Not actually allowed since WAPath is abstract (has abstract methods)
        path_json = WAPath(waypoints)

    Args:
        waypoints(np.ndarray): The waypoints that the path interpolates about or maintains
        **kwargs: Additional keyworded arguments.

    Raises:
        TypeError: the waypoints array type is not as expected
    """

    def __init__(self, waypoints, **kwargs):
        self._parameters = kwargs

        # Check points type and shape
        if isinstance(waypoints, list):
            waypoints = np.array(waypoints)
        elif not isinstance(waypoints, np.ndarray):
            raise TypeError(
                'waypoints type is not recognized. List or NumPy array required.')

        if 3 not in waypoints.shape:
            raise ValueError(
                f'waypoints shape is {waypoints.shape}, expected (n, 3) or (3, n).')

        self._waypoints = waypoints
        self._points = waypoints
        self._d_points = None

        self._is_closed = False if 'is_closed' not in kwargs else bool(kwargs['is_closed'])
        self._vis_properties = dict() if 'vis_properties' not in kwargs else kwargs['vis_properties']

    def get_points(self, der=0) -> np.ndarray:
        """Get the points for this path

        Args:
            der(int): derivative to grab. Defaults to 0 (just the points).

        Return:
            np.ndarray: The points array

        Raises:
            ValueError: If der is not a supported value
        """
        if der == 0:
            return self._points
        elif der == 1:
            return self._d_points
        else:
            raise ValueError(f'der value of {der} is not supported.')

    def get_waypoints(self) -> np.ndarray:
        """Get the waypoints for this path

        Return:
            np.ndarray: The waypoints array
        """
        return self._waypoints

    def is_closed(self) -> bool:
        """Get whether the path is closed

        Returns:
            bool: Is the path closed?
        """
        return self._is_closed

    def get_parameters(self) -> dict:
        """Get the parameters passed in to the this function.

        Track objects essentially copy other paths, so we want to keep the parameters for later

        Returns:
            dict: The saved parameteres
        """
        return self._parameters

    def set_vis_properties(self, vis_properties: dict):
        """Set the visual properties for this path.

        The visual properties are used in :meth:`~plot`.

        Args:
            vis_properties (dict): The visual properties to apply to this path.

        Raises:
            TypeError: If ``vis_properties`` is not a ``dict``
        """
        if not isinstance(vis_properties, dict):
            raise TypeError(f"'vis_properties' was expected to be a {type(dict)}, but was {type(vis_properties)}.")

        self._vis_properties = vis_properties

    def get_vis_properties(self) -> dict:
        """Get the visual properties.

        Python will return a reference since it is not a primitive type. This means that if you change any values
        in the returned properties dictionary, it will also change the instance held by this class. This can be an
        alternative method to the :meth:`~set_vis_properties`.

        Returns:
            dict: The visual properties
        """
        return self._vis_properties

    @ abstractmethod
    def calc_closest_point(self, pos: WAVector, return_idx: bool = False) -> WAVector:
        """Calculate the closest point on the path from the passed position

        Args:
            pos(WAVector): the position to find the closest point on the path to
            return_idx(bool, optional): return the index of the point with respect to the self._points array

        Returns:
            WAVector: the closest point on the path
            int(optional): the index of the point on the path
        """
        pass

    @ abstractmethod
    def plot(self, *args, show: bool = True, **kwargs):
        """Plot the path

        Args:
            *args: Positional arguments that are passed directly to the plotter
            show(bool, optional): show the plot window. Defaults to True.
            **kwargs: Keyworded arguments passed to the plotter
        """
        pass


class WASplinePath(WAPath):
    """Spline path implemented with SciPy's splprep and splev methods

    Args:
        waypoints(np.ndarray): the waypoints to fit the spline to
        num_points(int, optional): number of points to interpolate. Defaults to 100.
        smoothness(float, optional): how fit to each point the spline should be. will hit all points by default. Defaults to 0.0.
        is_closed(bool, optional): Is the path a closed loop. Defaults to True.

    Raises:
        TypeError: the waypoints array type is not as expected
    """

    def __init__(self, waypoints, **kwargs):
        # Check inputs
        allowed_args = {'num_points': 100, 'smoothness': 0.0, 'is_closed': False}  # noqa
        for key, value in allowed_args.items():
            if key in kwargs:
                value = kwargs[key]
            setattr(self, '_' + key, value)

        super().__init__(waypoints, **kwargs)

        # Check if the path is actuall closed
        if self._is_closed and not np.array_equal(waypoints[0], waypoints[-1]):
            warnings.warn(
                "is_closed has been set to True, but the first and last waypoints are not equal. Setting is_closed to False.", RuntimeWarning, stacklevel=100)
            self._is_closed = False

        # Interpolate the path
        tck, u = splprep(self._waypoints.T, s=self._smoothness, per=self._is_closed)  # noqa
        u_new = np.linspace(u.min(), u.max(), self._num_points)

        # Evaluate the interpolation to get values
        self._x, self._y, self._z = splev(u_new, tck, der=0)  # position
        self._dx, self._dy, self._dz = splev(u_new, tck, der=1)  # first derivative # noqa
        self._ddx, self._ddy, self._ddz = splev(u_new, tck, der=2)  # second derivative # noqa

        # store the points for later
        self._points = np.column_stack((self._x, self._y, self._z))
        self._d_points = np.column_stack((self._dx, self._dy, self._dz))

        # Variables for tracking path
        self._last_index = None

    def calc_closest_point(self, pos: WAVector, return_idx: bool = False) -> (WAVector, int):
        dist = cdist(self._points, [pos])
        idx, = np.argmin(dist, axis=0)

        pos = WAVector([self._x[idx], self._y[idx], self._z[idx]])
        if return_idx:
            return pos, idx
        return pos

    def plot(self, *args, show=True, ignore_vis_properties=False, **kwargs):
        """Plot the path in matplotlib.

        Args:
            args: Positional arguments passed directly to matplotlib
            show(bool, optional): show the plot window. Defaults to True.
            ignore_vis_properties (bool, optional): If True, ignore the visual properties set through :meth:`~set_vis_properties`. If False, dict passed to :meth:`~set_vis_properties` will be passed as keyworded argements to matplotlib.
            kwargs: Keyworded arguments passed directly to matplotlib.
        """
        plt.plot(self._x, self._y, *args, **self._vis_properties, **kwargs)
        if show:
            plt.show()

    def calc_length_cummulative(self) -> np.ndarray:
        """Get the cummulative distance along the path

        Returns:
            np.ndarray: Cummulative distance along the path
        """
        return calc_path_length_cummulative(self._x, self._y)

    def calc_curvature(self) -> np.ndarray:
        """Get the curvature at each point on the path

        Returns:
            np.ndarray: Curvature at each point on the path
        """
        return calc_path_curvature(self._dx, self._dy, self._ddx, self._ddy)
